"""
Implementation of the XDG Menu Specification Version 1.0.draft-1
http://standards.freedesktop.org/menu-spec/

Example code:

from xdg.Menu import parse, Menu, MenuEntry

def print_menu(menu, tab=0):
  for submenu in menu.Entries:
    if isinstance(submenu, Menu):
      print ("\t" * tab) + unicode(submenu)
      print_menu(submenu, tab+1)
    elif isinstance(submenu, MenuEntry):
      print ("\t" * tab) + unicode(submenu.DesktopEntry)

print_menu(parse())
"""

import locale, os, xml.dom.minidom
import subprocess

from xdg.BaseDirectory import xdg_data_dirs, xdg_config_dirs
from xdg.DesktopEntry import DesktopEntry
from xdg.Exceptions import ParsingError, ValidationError, debug
from xdg.util import PY3

import xdg.Locale
import xdg.Config

ELEMENT_NODE = xml.dom.Node.ELEMENT_NODE

def _strxfrm(s):
    """Wrapper around locale.strxfrm that accepts unicode strings on Python 2.
    
    See Python bug #2481.
    """
    if (not PY3) and isinstance(s, unicode):
        s = s.encode('utf-8')
    return locale.strxfrm(s)

class Menu:
    """Menu containing sub menus under menu.Entries

	Contains both Menu and MenuEntry items.
    """
    def __init__(self):
        # Public stuff
        self.Name = ""
        self.Directory = None
        self.Entries = []
        self.Doc = ""
        self.Filename = ""
        self.Depth = 0
        self.Parent = None
        self.NotInXml = False

        # Can be one of Deleted/NoDisplay/Hidden/Empty/NotShowIn or True
        self.Show = True
        self.Visible = 0

        # Private stuff, only needed for parsing
        self.AppDirs = []
        self.DefaultLayout = None
        self.Deleted = "notset"
        self.Directories = []
        self.DirectoryDirs = []
        self.Layout = None
        self.MenuEntries = []
        self.Moves = []
        self.OnlyUnallocated = "notset"
        self.Rules = []
        self.Submenus = []

    def __str__(self):
        return self.Name

    def __add__(self, other):
        for dir in other.AppDirs:
            self.AppDirs.append(dir)

        for dir in other.DirectoryDirs:
            self.DirectoryDirs.append(dir)

        for directory in other.Directories:
            self.Directories.append(directory)

        if other.Deleted != "notset":
            self.Deleted = other.Deleted

        if other.OnlyUnallocated != "notset":
            self.OnlyUnallocated = other.OnlyUnallocated

        if other.Layout:
            self.Layout = other.Layout

        if other.DefaultLayout:
            self.DefaultLayout = other.DefaultLayout

        for rule in other.Rules:
            self.Rules.append(rule)

        for move in other.Moves:
            self.Moves.append(move)

        for submenu in other.Submenus:
            self.addSubmenu(submenu)

        return self

    # FIXME: Performance: cache getName()
    def __cmp__(self, other):
        return locale.strcoll(self.getName(), other.getName())
    
    def _key(self):
        """Key function for locale-aware sorting."""
        return _strxfrm(self.getName())
    
    def __lt__(self, other):
        try:
            other = other._key()
        except AttributeError:
            pass
        return self._key() < other

    def __eq__(self, other):
        try:
            return self.Name == unicode(other)
        except NameError:  # unicode() becomes str() in Python 3
            return self.Name == str(other)

    """ PUBLIC STUFF """
    def getEntries(self, hidden=False):
        """Interator for a list of Entries visible to the user."""
        for entry in self.Entries:
            if hidden == True:
                yield entry
            elif entry.Show == True:
                yield entry

    # FIXME: Add searchEntry/seaqrchMenu function
    # search for name/comment/genericname/desktopfileide
    # return multiple items

    def getMenuEntry(self, desktopfileid, deep = False):
        """Searches for a MenuEntry with a given DesktopFileID."""
        for menuentry in self.MenuEntries:
            if menuentry.DesktopFileID == desktopfileid:
                return menuentry
        if deep == True:
            for submenu in self.Submenus:
                submenu.getMenuEntry(desktopfileid, deep)

    def getMenu(self, path):
        """Searches for a Menu with a given path."""
        array = path.split("/", 1)
        for submenu in self.Submenus:
            if submenu.Name == array[0]:
                if len(array) > 1:
                    return submenu.getMenu(array[1])
                else:
                    return submenu

    def getPath(self, org=False, toplevel=False):
        """Returns this menu's path in the menu structure."""
        parent = self
        names=[]
        while 1:
            if org:
                names.append(parent.Name)
            else:
                names.append(parent.getName())
            if parent.Depth > 0:
                parent = parent.Parent
            else:
                break
        names.reverse()
        path = ""
        if toplevel == False:
            names.pop(0)
        for name in names:
            path = os.path.join(path, name)
        return path

    def getName(self):
        """Returns the menu's localised name."""
        try:
            return self.Directory.DesktopEntry.getName()
        except AttributeError:
            return self.Name

    def getGenericName(self):
        """Returns the menu's generic name."""
        try:
            return self.Directory.DesktopEntry.getGenericName()
        except AttributeError:
            return ""

    def getComment(self):
        """Returns the menu's comment text."""
        try:
            return self.Directory.DesktopEntry.getComment()
        except AttributeError:
            return ""

    def getIcon(self):
        """Returns the menu's icon, filename or simple name"""
        try:
            return self.Directory.DesktopEntry.getIcon()
        except AttributeError:
            return ""

    """ PRIVATE STUFF """
    def addSubmenu(self, newmenu):
        for submenu in self.Submenus:
            if submenu == newmenu:
                submenu += newmenu
                break
        else:
            self.Submenus.append(newmenu)
            newmenu.Parent = self
            newmenu.Depth = self.Depth + 1

class Move:
    "A move operation"
    def __init__(self, node=None):
        if node:
            self.parseNode(node)
        else:
            self.Old = ""
            self.New = ""

    def __cmp__(self, other):
        return cmp(self.Old, other.Old)

    def parseNode(self, node):
        for child in node.childNodes:
            if child.nodeType == ELEMENT_NODE:
                if child.tagName == "Old":
                    try:
                        self.parseOld(child.childNodes[0].nodeValue)
                    except IndexError:
                        raise ValidationError('Old cannot be empty', '??')                                            
                elif child.tagName == "New":
                    try:
                        self.parseNew(child.childNodes[0].nodeValue)
                    except IndexError:
                        raise ValidationError('New cannot be empty', '??')                                            

    def parseOld(self, value):
        self.Old = value
    def parseNew(self, value):
        self.New = value


class Layout:
    "Menu Layout class"
    def __init__(self, node=None):
        self.order = []
        if node:
            self.show_empty = node.getAttribute("show_empty") or "false"
            self.inline = node.getAttribute("inline") or "false"
            self.inline_limit = node.getAttribute("inline_limit") or 4
            self.inline_header = node.getAttribute("inline_header") or "true"
            self.inline_alias = node.getAttribute("inline_alias") or "false"
            self.inline_limit = int(self.inline_limit)
            self.parseNode(node)
        else:
            self.show_empty = "false"
            self.inline = "false"
            self.inline_limit = 4
            self.inline_header = "true"
            self.inline_alias = "false"
            self.order.append(["Merge", "menus"])
            self.order.append(["Merge", "files"])

    def parseNode(self, node):
        for child in node.childNodes:
            if child.nodeType == ELEMENT_NODE:
                if child.tagName == "Menuname":
                    try:
                        self.parseMenuname(
                            child.childNodes[0].nodeValue,
                            child.getAttribute("show_empty") or "false",
                            child.getAttribute("inline") or "false",
                            child.getAttribute("inline_limit") or 4,
                            child.getAttribute("inline_header") or "true",
                            child.getAttribute("inline_alias") or "false" )
                    except IndexError:
                        raise ValidationError('Menuname cannot be empty', "")
                elif child.tagName == "Separator":
                    self.parseSeparator()
                elif child.tagName == "Filename":
                    try:
                        self.parseFilename(child.childNodes[0].nodeValue)
                    except IndexError:
                        raise ValidationError('Filename cannot be empty', "")
                elif child.tagName == "Merge":
                    self.parseMerge(child.getAttribute("type") or "all")

    def parseMenuname(self, value, empty="false", inline="false", inline_limit=4, inline_header="true", inline_alias="false"):
        self.order.append(["Menuname", value, empty, inline, inline_limit, inline_header, inline_alias])
        self.order[-1][4] = int(self.order[-1][4])

    def parseSeparator(self):
        self.order.append(["Separator"])

    def parseFilename(self, value):
        self.order.append(["Filename", value])

    def parseMerge(self, type="all"):
        self.order.append(["Merge", type])


class Rule:
    "Inlcude / Exclude Rules Class"
    def __init__(self, type, node=None):
        # Type is Include or Exclude
        self.Type = type
        # Rule is a python expression
        self.Rule = ""

        # Private attributes, only needed for parsing
        self.Depth = 0
        self.Expr = [ "or" ]
        self.New = True

        # Begin parsing
        if node:
            self.parseNode(node)

    def __str__(self):
        return self.Rule
    
    def do(self, menuentries, type, run):
        for menuentry in menuentries:
            if run == 2 and ( menuentry.MatchedInclude == True \
            or menuentry.Allocated == True ):
                continue
            elif eval(self.Rule):
                if type == "Include":
                    menuentry.Add = True
                    menuentry.MatchedInclude = True
                else:
                    menuentry.Add = False
        return menuentries

    def parseNode(self, node):
        for child in node.childNodes:
            if child.nodeType == ELEMENT_NODE:
                if child.tagName == 'Filename':
                    try:
                        self.parseFilename(child.childNodes[0].nodeValue)
                    except IndexError:
                        raise ValidationError('Filename cannot be empty', "???")
                elif child.tagName == 'Category':
                    try:
                        self.parseCategory(child.childNodes[0].nodeValue)
                    except IndexError:
                        raise ValidationError('Category cannot be empty', "???")
                elif child.tagName == 'All':
                    self.parseAll()
                elif child.tagName == 'And':
                    self.parseAnd(child)
                elif child.tagName == 'Or':
                    self.parseOr(child)
                elif child.tagName == 'Not':
                    self.parseNot(child)

    def parseNew(self, set=True):
        if not self.New:
            self.Rule += " " + self.Expr[self.Depth] + " "
        if not set:
            self.New = True
        elif set:
            self.New = False

    def parseFilename(self, value):
        self.parseNew()
        self.Rule += "menuentry.DesktopFileID == '%s'" % value.strip().replace("\\", r"\\").replace("'", r"\'")

    def parseCategory(self, value):
        self.parseNew()
        self.Rule += "'%s' in menuentry.Categories" % value.strip()

    def parseAll(self):
        self.parseNew()
        self.Rule += "True"

    def parseAnd(self, node):
        self.parseNew(False)
        self.Rule += "("
        self.Depth += 1
        self.Expr.append("and")
        self.parseNode(node)
        self.Depth -= 1
        self.Expr.pop()
        self.Rule += ")"

    def parseOr(self, node):
        self.parseNew(False)
        self.Rule += "("
        self.Depth += 1
        self.Expr.append("or")
        self.parseNode(node)
        self.Depth -= 1
        self.Expr.pop()
        self.Rule += ")"

    def parseNot(self, node):
        self.parseNew(False)
        self.Rule += "not ("
        self.Depth += 1
        self.Expr.append("or")
        self.parseNode(node)
        self.Depth -= 1
        self.Expr.pop()
        self.Rule += ")"


class MenuEntry:
    "Wrapper for 'Menu Style' Desktop Entries"
    def __init__(self, filename, dir="", prefix=""):
        # Create entry
        self.DesktopEntry = DesktopEntry(os.path.join(dir,filename))
        self.setAttributes(filename, dir, prefix)

        # Can be one of Deleted/Hidden/Empty/NotShowIn/NoExec or True
        self.Show = True

        # Semi-Private
        self.Original = None
        self.Parents = []

        # Private Stuff
        self.Allocated = False
        self.Add = False
        self.MatchedInclude = False

        # Caching
        self.Categories = self.DesktopEntry.getCategories()

    def save(self):
        """Save any changes to the desktop entry."""
        if self.DesktopEntry.tainted == True:
            self.DesktopEntry.write()

    def getDir(self):
        """Return the directory containing the desktop entry file."""
        return self.DesktopEntry.filename.replace(self.Filename, '')

    def getType(self):
        """Return the type of MenuEntry, System/User/Both"""
        if xdg.Config.root_mode == False:
            if self.Original:
                return "Both"
            elif xdg_data_dirs[0] in self.DesktopEntry.filename:
                return "User"
            else:
                return "System"
        else:
            return "User"

    def setAttributes(self, filename, dir="", prefix=""):
        self.Filename = filename
        self.Prefix = prefix
        self.DesktopFileID = os.path.join(prefix,filename).replace("/", "-")

        if not os.path.isabs(self.DesktopEntry.filename):
            self.__setFilename()

    def updateAttributes(self):
        if self.getType() == "System":
            self.Original = MenuEntry(self.Filename, self.getDir(), self.Prefix)
            self.__setFilename()

    def __setFilename(self):
        if xdg.Config.root_mode == False:
            path = xdg_data_dirs[0]
        else:
            path= xdg_data_dirs[1]

        if self.DesktopEntry.getType() == "Application":
            dir = os.path.join(path, "applications")
        else:
            dir = os.path.join(path, "desktop-directories")

        self.DesktopEntry.filename = os.path.join(dir, self.Filename)

    def __cmp__(self, other):
        return locale.strcoll(self.DesktopEntry.getName(), other.DesktopEntry.getName())
    
    def _key(self):
        """Key function for locale-aware sorting."""
        return _strxfrm(self.DesktopEntry.getName())
    
    def __lt__(self, other):
        try:
            other = other._key()
        except AttributeError:
            pass
        return self._key() < other
            

    def __eq__(self, other):
        if self.DesktopFileID == str(other):
            return True
        else:
            return False

    def __repr__(self):
        return self.DesktopFileID


class Separator:
    "Just a dummy class for Separators"
    def __init__(self, parent):
        self.Parent = parent
        self.Show = True


class Header:
    "Class for Inline Headers"
    def __init__(self, name, generic_name, comment):
        self.Name = name
        self.GenericName = generic_name
        self.Comment = comment

    def __str__(self):
        return self.Name


tmp = {}

def __getFileName(filename):
    dirs = xdg_config_dirs[:]
    if xdg.Config.root_mode == True:
        dirs.pop(0)

    for dir in dirs:
        menuname = os.path.join (dir, "menus" , filename)
        if os.path.isdir(dir) and os.path.isfile(menuname):
            return menuname

def parse(filename=None):
    """Load an applications.menu file.
    
    filename : str, optional
      The default is ``$XDG_CONFIG_DIRS/menus/${XDG_MENU_PREFIX}applications.menu``.
    """
    # convert to absolute path
    if filename and not os.path.isabs(filename):
        filename = __getFileName(filename)

    # use default if no filename given
    if not filename: 
        candidate = os.environ.get('XDG_MENU_PREFIX', '') + "applications.menu"
        filename = __getFileName(candidate)
        
    if not filename:
        raise ParsingError('File not found', "/etc/xdg/menus/%s" % candidate)

    # check if it is a .menu file
    if not os.path.splitext(filename)[1] == ".menu":
        raise ParsingError('Not a .menu file', filename)

    # create xml parser
    try:
        doc = xml.dom.minidom.parse(filename)
    except xml.parsers.expat.ExpatError:
        raise ParsingError('Not a valid .menu file', filename)

    # parse menufile
    tmp["Root"] = ""
    tmp["mergeFiles"] = []
    tmp["DirectoryDirs"] = []
    tmp["cache"] = MenuEntryCache()

    __parse(doc, filename, tmp["Root"])
    __parsemove(tmp["Root"])
    __postparse(tmp["Root"])

    tmp["Root"].Doc = doc
    tmp["Root"].Filename = filename

    # generate the menu
    __genmenuNotOnlyAllocated(tmp["Root"])
    __genmenuOnlyAllocated(tmp["Root"])

    # and finally sort
    sort(tmp["Root"])

    return tmp["Root"]


def __parse(node, filename, parent=None):
    for child in node.childNodes:
        if child.nodeType == ELEMENT_NODE:
            if child.tagName == 'Menu':
                __parseMenu(child, filename, parent)
            elif child.tagName == 'AppDir':
                try:
                    __parseAppDir(child.childNodes[0].nodeValue, filename, parent)
                except IndexError:
                    raise ValidationError('AppDir cannot be empty', filename)
            elif child.tagName == 'DefaultAppDirs':
                __parseDefaultAppDir(filename, parent)
            elif child.tagName == 'DirectoryDir':
                try:
                    __parseDirectoryDir(child.childNodes[0].nodeValue, filename, parent)
                except IndexError:
                    raise ValidationError('DirectoryDir cannot be empty', filename)
            elif child.tagName == 'DefaultDirectoryDirs':
                __parseDefaultDirectoryDir(filename, parent)
            elif child.tagName == 'Name' :
                try:
                    parent.Name = child.childNodes[0].nodeValue
                except IndexError:
                    raise ValidationError('Name cannot be empty', filename)
            elif child.tagName == 'Directory' :
                try:
                    parent.Directories.append(child.childNodes[0].nodeValue)
                except IndexError:
                    raise ValidationError('Directory cannot be empty', filename)
            elif child.tagName == 'OnlyUnallocated':
                parent.OnlyUnallocated = True
            elif child.tagName == 'NotOnlyUnallocated':
                parent.OnlyUnallocated = False
            elif child.tagName == 'Deleted':
                parent.Deleted = True
            elif child.tagName == 'NotDeleted':
                parent.Deleted = False
            elif child.tagName == 'Include' or child.tagName == 'Exclude':
                parent.Rules.append(Rule(child.tagName, child))
            elif child.tagName == 'MergeFile':
                try:
                    if child.getAttribute("type") == "parent":
                        __parseMergeFile("applications.menu", child, filename, parent)
                    else:
                        __parseMergeFile(child.childNodes[0].nodeValue, child, filename, parent)
                except IndexError:
                    raise ValidationError('MergeFile cannot be empty', filename)
            elif child.tagName == 'MergeDir':
                try:
                    __parseMergeDir(child.childNodes[0].nodeValue, child, filename, parent)
                except IndexError:
                    raise ValidationError('MergeDir cannot be empty', filename)
            elif child.tagName == 'DefaultMergeDirs':
                __parseDefaultMergeDirs(child, filename, parent)
            elif child.tagName == 'Move':
                parent.Moves.append(Move(child))
            elif child.tagName == 'Layout':
                if len(child.childNodes) > 1:
                    parent.Layout = Layout(child)
            elif child.tagName == 'DefaultLayout':
                if len(child.childNodes) > 1:
                    parent.DefaultLayout = Layout(child)
            elif child.tagName == 'LegacyDir':
                try:
                    __parseLegacyDir(child.childNodes[0].nodeValue, child.getAttribute("prefix"), filename, parent)
                except IndexError:
                    raise ValidationError('LegacyDir cannot be empty', filename)
            elif child.tagName == 'KDELegacyDirs':
                __parseKDELegacyDirs(filename, parent)

def __parsemove(menu):
    for submenu in menu.Submenus:
        __parsemove(submenu)

    # parse move operations
    for move in menu.Moves:
        move_from_menu = menu.getMenu(move.Old)
        if move_from_menu:
            move_to_menu = menu.getMenu(move.New)

            menus = move.New.split("/")
            oldparent = None
            while len(menus) > 0:
                if not oldparent:
                    oldparent = menu
                newmenu = oldparent.getMenu(menus[0])
                if not newmenu:
                    newmenu = Menu()
                    newmenu.Name = menus[0]
                    if len(menus) > 1:
                        newmenu.NotInXml = True
                    oldparent.addSubmenu(newmenu)
                oldparent = newmenu
                menus.pop(0)

            newmenu += move_from_menu
            move_from_menu.Parent.Submenus.remove(move_from_menu)

def __postparse(menu):
    # unallocated / deleted
    if menu.Deleted == "notset":
        menu.Deleted = False
    if menu.OnlyUnallocated == "notset":
        menu.OnlyUnallocated = False

    # Layout Tags
    if not menu.Layout or not menu.DefaultLayout:
        if menu.DefaultLayout:
            menu.Layout = menu.DefaultLayout
        elif menu.Layout:
            if menu.Depth > 0:
                menu.DefaultLayout = menu.Parent.DefaultLayout
            else:
                menu.DefaultLayout = Layout()
        else:
            if menu.Depth > 0:
                menu.Layout = menu.Parent.DefaultLayout
                menu.DefaultLayout = menu.Parent.DefaultLayout
            else:
                menu.Layout = Layout()
                menu.DefaultLayout = Layout()

    # add parent's app/directory dirs
    if menu.Depth > 0:
        menu.AppDirs = menu.Parent.AppDirs + menu.AppDirs
        menu.DirectoryDirs = menu.Parent.DirectoryDirs + menu.DirectoryDirs

    # remove duplicates
    menu.Directories = __removeDuplicates(menu.Directories)
    menu.DirectoryDirs = __removeDuplicates(menu.DirectoryDirs)
    menu.AppDirs = __removeDuplicates(menu.AppDirs)

    # go recursive through all menus
    for submenu in menu.Submenus:
        __postparse(submenu)

    # reverse so handling is easier
    menu.Directories.reverse()
    menu.DirectoryDirs.reverse()
    menu.AppDirs.reverse()

    # get the valid .directory file out of the list
    for directory in menu.Directories:
        for dir in menu.DirectoryDirs:
            if os.path.isfile(os.path.join(dir, directory)):
                menuentry = MenuEntry(directory, dir)
                if not menu.Directory:
                    menu.Directory = menuentry
                elif menuentry.getType() == "System":
                    if menu.Directory.getType() == "User":
                        menu.Directory.Original = menuentry
        if menu.Directory:
            break


# Menu parsing stuff
def __parseMenu(child, filename, parent):
    m = Menu()
    __parse(child, filename, m)
    if parent:
        parent.addSubmenu(m)
    else:
        tmp["Root"] = m

# helper function
def __check(value, filename, type):
    path = os.path.dirname(filename)

    if not os.path.isabs(value):
        value = os.path.join(path, value)

    value = os.path.abspath(value)

    if type == "dir" and os.path.exists(value) and os.path.isdir(value):
        return value
    elif type == "file" and os.path.exists(value) and os.path.isfile(value):
        return value
    else:
        return False

# App/Directory Dir Stuff
def __parseAppDir(value, filename, parent):
    value = __check(value, filename, "dir")
    if value:
        parent.AppDirs.append(value)

def __parseDefaultAppDir(filename, parent):
    for dir in reversed(xdg_data_dirs):
        __parseAppDir(os.path.join(dir, "applications"), filename, parent)

def __parseDirectoryDir(value, filename, parent):
    value = __check(value, filename, "dir")
    if value:
        parent.DirectoryDirs.append(value)

def __parseDefaultDirectoryDir(filename, parent):
    for dir in reversed(xdg_data_dirs):
        __parseDirectoryDir(os.path.join(dir, "desktop-directories"), filename, parent)

# Merge Stuff
def __parseMergeFile(value, child, filename, parent):
    if child.getAttribute("type") == "parent":
        for dir in xdg_config_dirs:
            rel_file = filename.replace(dir, "").strip("/")
            if rel_file != filename:
                for p in xdg_config_dirs:
                    if dir == p:
                        continue
                    if os.path.isfile(os.path.join(p,rel_file)):
                        __mergeFile(os.path.join(p,rel_file),child,parent)
                        break
    else:
        value = __check(value, filename, "file")
        if value:
            __mergeFile(value, child, parent)

def __parseMergeDir(value, child, filename, parent):
    value = __check(value, filename, "dir")
    if value:
        for item in os.listdir(value):
            try:
                if os.path.splitext(item)[1] == ".menu":
                    __mergeFile(os.path.join(value, item), child, parent)
            except UnicodeDecodeError:
                continue

def __parseDefaultMergeDirs(child, filename, parent):
    basename = os.path.splitext(os.path.basename(filename))[0]
    for dir in reversed(xdg_config_dirs):
        __parseMergeDir(os.path.join(dir, "menus", basename + "-merged"), child, filename, parent)

def __mergeFile(filename, child, parent):
    # check for infinite loops
    if filename in tmp["mergeFiles"]:
        if debug:
            raise ParsingError('Infinite MergeFile loop detected', filename)
        else:
            return

    tmp["mergeFiles"].append(filename)

    # load file
    try:
        doc = xml.dom.minidom.parse(filename)
    except IOError:
        if debug:
            raise ParsingError('File not found', filename)
        else:
            return
    except xml.parsers.expat.ExpatError:
        if debug:
            raise ParsingError('Not a valid .menu file', filename)
        else:
            return

    # append file
    for child in doc.childNodes:
        if child.nodeType == ELEMENT_NODE:
            __parse(child,filename,parent)
            break

# Legacy Dir Stuff
def __parseLegacyDir(dir, prefix, filename, parent):
    m = __mergeLegacyDir(dir,prefix,filename,parent)
    if m:
        parent += m

def __mergeLegacyDir(dir, prefix, filename, parent):
    dir = __check(dir,filename,"dir")
    if dir and dir not in tmp["DirectoryDirs"]:
        tmp["DirectoryDirs"].append(dir)

        m = Menu()
        m.AppDirs.append(dir)
        m.DirectoryDirs.append(dir)
        m.Name = os.path.basename(dir)
        m.NotInXml = True

        for item in os.listdir(dir):
            try:
                if item == ".directory":
                    m.Directories.append(item)
                elif os.path.isdir(os.path.join(dir,item)):
                    m.addSubmenu(__mergeLegacyDir(os.path.join(dir,item), prefix, filename, parent))
            except UnicodeDecodeError:
                continue

        tmp["cache"].addMenuEntries([dir],prefix, True)
        menuentries = tmp["cache"].getMenuEntries([dir], False)

        for menuentry in menuentries:
            categories = menuentry.Categories
            if len(categories) == 0:
                r = Rule("Include")
                r.parseFilename(menuentry.DesktopFileID)
                m.Rules.append(r)
            if not dir in parent.AppDirs:
                categories.append("Legacy")
                menuentry.Categories = categories

        return m

def __parseKDELegacyDirs(filename, parent):
    try:
        proc = subprocess.Popen(['kde-config', '--path', 'apps'],
                                stdout=subprocess.PIPE, universal_newlines=True)
        output = proc.communicate()[0].splitlines()
    except OSError:
        # If kde-config doesn't exist, ignore this.
        return
    
    try:
        for dir in output[0].split(":"):
            __parseLegacyDir(dir,"kde", filename, parent)
    except IndexError:
        pass

# remove duplicate entries from a list
def __removeDuplicates(list):
    set = {}
    list.reverse()
    list = [set.setdefault(e,e) for e in list if e not in set]
    list.reverse()
    return list

# Finally generate the menu
def __genmenuNotOnlyAllocated(menu):
    for submenu in menu.Submenus:
        __genmenuNotOnlyAllocated(submenu)

    if menu.OnlyUnallocated == False:
        tmp["cache"].addMenuEntries(menu.AppDirs)
        menuentries = []
        for rule in menu.Rules:
            menuentries = rule.do(tmp["cache"].getMenuEntries(menu.AppDirs), rule.Type, 1)
        for menuentry in menuentries:
            if menuentry.Add == True:
                menuentry.Parents.append(menu)
                menuentry.Add = False
                menuentry.Allocated = True
                menu.MenuEntries.append(menuentry)

def __genmenuOnlyAllocated(menu):
    for submenu in menu.Submenus:
        __genmenuOnlyAllocated(submenu)

    if menu.OnlyUnallocated == True:
        tmp["cache"].addMenuEntries(menu.AppDirs)
        menuentries = []
        for rule in menu.Rules:
            menuentries = rule.do(tmp["cache"].getMenuEntries(menu.AppDirs), rule.Type, 2)
        for menuentry in menuentries:
            if menuentry.Add == True:
                menuentry.Parents.append(menu)
            #   menuentry.Add = False
            #   menuentry.Allocated = True
                menu.MenuEntries.append(menuentry)

# And sorting ...
def sort(menu):
    menu.Entries = []
    menu.Visible = 0

    for submenu in menu.Submenus:
        sort(submenu)

    tmp_s = []
    tmp_e = []

    for order in menu.Layout.order:
        if order[0] == "Filename":
            tmp_e.append(order[1])
        elif order[0] == "Menuname":
            tmp_s.append(order[1])
    
    for order in menu.Layout.order:
        if order[0] == "Separator":
            separator = Separator(menu)
            if len(menu.Entries) > 0 and isinstance(menu.Entries[-1], Separator):
                separator.Show = False
            menu.Entries.append(separator)
        elif order[0] == "Filename":
            menuentry = menu.getMenuEntry(order[1])
            if menuentry:
                menu.Entries.append(menuentry)
        elif order[0] == "Menuname":
            submenu = menu.getMenu(order[1])
            if submenu:
                __parse_inline(submenu, menu)
        elif order[0] == "Merge":
            if order[1] == "files" or order[1] == "all":
                menu.MenuEntries.sort()
                for menuentry in menu.MenuEntries:
                    if menuentry not in tmp_e:
                        menu.Entries.append(menuentry)
            elif order[1] == "menus" or order[1] == "all":
                menu.Submenus.sort()
                for submenu in menu.Submenus:
                    if submenu.Name not in tmp_s:
                        __parse_inline(submenu, menu)

    # getHidden / NoDisplay / OnlyShowIn / NotOnlyShowIn / Deleted / NoExec
    for entry in menu.Entries:
        entry.Show = True
        menu.Visible += 1
        if isinstance(entry, Menu):
            if entry.Deleted == True:
                entry.Show = "Deleted"
                menu.Visible -= 1
            elif isinstance(entry.Directory, MenuEntry):
                if entry.Directory.DesktopEntry.getNoDisplay() == True:
                    entry.Show = "NoDisplay"
                    menu.Visible -= 1
                elif entry.Directory.DesktopEntry.getHidden() == True:
                    entry.Show = "Hidden"
                    menu.Visible -= 1
        elif isinstance(entry, MenuEntry):
            if entry.DesktopEntry.getNoDisplay() == True:
                entry.Show = "NoDisplay"
                menu.Visible -= 1
            elif entry.DesktopEntry.getHidden() == True:
                entry.Show = "Hidden"
                menu.Visible -= 1
            elif entry.DesktopEntry.getTryExec() and not __try_exec(entry.DesktopEntry.getTryExec()):
                entry.Show = "NoExec"
                menu.Visible -= 1
            elif xdg.Config.windowmanager:
                if ( entry.DesktopEntry.getOnlyShowIn() != [] and xdg.Config.windowmanager not in entry.DesktopEntry.getOnlyShowIn() ) \
                or xdg.Config.windowmanager in entry.DesktopEntry.getNotShowIn():
                    entry.Show = "NotShowIn"
                    menu.Visible -= 1
        elif isinstance(entry,Separator):
            menu.Visible -= 1

    # remove separators at the beginning and at the end
    if len(menu.Entries) > 0:
        if isinstance(menu.Entries[0], Separator):
            menu.Entries[0].Show = False
    if len(menu.Entries) > 1:
        if isinstance(menu.Entries[-1], Separator):
            menu.Entries[-1].Show = False

    # show_empty tag
    for entry in menu.Entries[:]:
        if isinstance(entry, Menu) and entry.Layout.show_empty == "false" and entry.Visible == 0:
            entry.Show = "Empty"
            menu.Visible -= 1
            if entry.NotInXml == True:
                menu.Entries.remove(entry)

def __try_exec(executable):
    paths = os.environ['PATH'].split(os.pathsep)
    if not os.path.isfile(executable):
        for p in paths:
            f = os.path.join(p, executable)
            if os.path.isfile(f):
                if os.access(f, os.X_OK):
                    return True
    else:
        if os.access(executable, os.X_OK):
            return True
    return False

# inline tags
def __parse_inline(submenu, menu):
    if submenu.Layout.inline == "true":
        if len(submenu.Entries) == 1 and submenu.Layout.inline_alias == "true":
            menuentry = submenu.Entries[0]
            menuentry.DesktopEntry.set("Name", submenu.getName(), locale = True)
            menuentry.DesktopEntry.set("GenericName", submenu.getGenericName(), locale = True)
            menuentry.DesktopEntry.set("Comment", submenu.getComment(), locale = True)
            menu.Entries.append(menuentry)
        elif len(submenu.Entries) <= submenu.Layout.inline_limit or submenu.Layout.inline_limit == 0:
            if submenu.Layout.inline_header == "true":
                header = Header(submenu.getName(), submenu.getGenericName(), submenu.getComment())
                menu.Entries.append(header)
            for entry in submenu.Entries:
                menu.Entries.append(entry)
        else:
            menu.Entries.append(submenu)
    else:
        menu.Entries.append(submenu)

class MenuEntryCache:
    "Class to cache Desktop Entries"
    def __init__(self):
        self.cacheEntries = {}
        self.cacheEntries['legacy'] = []
        self.cache = {}

    def addMenuEntries(self, dirs, prefix="", legacy=False):
        for dir in dirs:
            if not dir in self.cacheEntries:
                self.cacheEntries[dir] = []
                self.__addFiles(dir, "", prefix, legacy)

    def __addFiles(self, dir, subdir, prefix, legacy):
        for item in os.listdir(os.path.join(dir,subdir)):
            if os.path.splitext(item)[1] == ".desktop":
                try:
                    menuentry = MenuEntry(os.path.join(subdir,item), dir, prefix)
                except ParsingError:
                    continue

                self.cacheEntries[dir].append(menuentry)
                if legacy == True:
                    self.cacheEntries['legacy'].append(menuentry)
            elif os.path.isdir(os.path.join(dir,subdir,item)) and legacy == False:
                self.__addFiles(dir, os.path.join(subdir,item), prefix, legacy)

    def getMenuEntries(self, dirs, legacy=True):
        list = []
        ids = []
        # handle legacy items
        appdirs = dirs[:]
        if legacy == True:
            appdirs.append("legacy")
        # cache the results again
        key = "".join(appdirs)
        try:
            return self.cache[key]
        except KeyError:
            pass
        for dir in appdirs:
            for menuentry in self.cacheEntries[dir]:
                try:
                    if menuentry.DesktopFileID not in ids:
                        ids.append(menuentry.DesktopFileID)
                        list.append(menuentry)
                    elif menuentry.getType() == "System":
                    # FIXME: This is only 99% correct, but still...
                        i = list.index(menuentry)
                        e = list[i]
                        if e.getType() == "User":
                            e.Original = menuentry
                except UnicodeDecodeError:
                    continue
        self.cache[key] = list
        return list



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib import patterns, config
from taskcoachlib.domain import task, effort, date, category
from taskcoachlib.thirdparty.pubsub import pub
from unittests import asserts
import test
import wx


class EffortTest(test.TestCase, asserts.Mixin):
    def setUp(self):
        task.Task.settings = config.Settings(load=False)
        self.task = task.Task()
        self.effort = effort.Effort(self.task, start=date.DateTime(2004, 1, 1),
            stop=date.DateTime(2004, 1, 2))
        self.task.addEffort(self.effort)
        self.events = []
    
    def onEvent(self, event):
        self.events.append(event)
        
    def testId(self):
        self.assert_(self.effort.id() is not None)

    def testStatus(self):
        self.assertEqual(self.effort.getStatus(), self.effort.STATUS_NEW)

    def testCreate(self):
        self.assertEqual(self.task, self.effort.task())
        self.assertEqual('', self.effort.description())
        
    def testStr(self):
        self.assertEqual('Effort(%s, %s, %s)' % (self.effort.task(), 
            self.effort.getStart(), self.effort.getStop()), str(self.effort))
        
    def testDuration(self):
        self.assertEqual(date.TimeDelta(days=1), self.effort.duration())
        
    def testForegroundColor(self):
        self.task.setForegroundColor(wx.RED)
        self.assertEqual(wx.RED, self.effort.foregroundColor())

    def testBackgroundColor(self):
        self.task.setBackgroundColor(wx.RED)
        self.assertEqual(wx.RED, self.effort.backgroundColor())
        
    def testFont(self):
        self.task.setFont(wx.SWISS_FONT)
        self.assertEqual(wx.SWISS_FONT, self.effort.font())

    def testNotificationForSetStart(self):
        events = []
        
        def onEvent(newValue, sender):
            events.append((newValue, sender))
            
        pub.subscribe(onEvent, effort.Effort.startChangedEventType())
        start = date.DateTime.now()
        self.effort.setStart(start)
        self.assertEqual([(start, self.effort)], events)
        
    def testNotificationForSetStop(self):
        events = []
        
        def onEvent(newValue, sender):
            events.append((newValue, sender))
            
        pub.subscribe(onEvent, effort.Effort.stopChangedEventType())
        stop = date.DateTime.now()
        self.effort.setStop(stop)
        self.assertEqual([(stop, self.effort)], events)
        
    def testNoNotificationForSetStopWhenNewStopEqualsOldStop(self):
        events = []
        
        def onEvent(newValue, sender):
            events.append((newValue, sender))
            
        pub.subscribe(onEvent, effort.Effort.stopChangedEventType())
        self.effort.setStop(self.effort.getStop())
        self.failIf(events)

    def testDurationNotificationForSetStart(self):
        events = []
        
        def onEvent(newValue, sender):
            events.append((newValue, sender))
            
        pub.subscribe(onEvent, effort.Effort.durationChangedEventType())
        start = date.DateTime.now()
        self.effort.setStart(start)
        self.assertEqual([(self.effort.duration(), self.effort)], events)

    def testDurationNotificationForSetStop(self):
        events = []
        
        def onEvent(newValue, sender):
            events.append((newValue, sender))
            
        pub.subscribe(onEvent, effort.Effort.durationChangedEventType())
        self.effort.setStop(date.DateTime.now())
        self.assertEqual([(self.effort.duration(), self.effort)], events)

    def testNotificationForSetDescription(self):
        patterns.Publisher().registerObserver(self.onEvent,
            eventType=effort.Effort.descriptionChangedEventType())
        self.effort.setDescription('description')
        self.assertEqual('description', self.events[0].value())

    def testNotificationForSetTask(self):
        events = []
        
        def onEvent(newValue, sender):
            events.append((newValue, sender))
            
        pub.subscribe(onEvent, effort.Effort.taskChangedEventType())
        task2 = task.Task()
        self.effort.setTask(task2)
        self.assertEqual([(task2, self.effort)], events)

    def testNotificationForStartTracking(self):
        events = []
        
        def onEvent(newValue, sender):
            events.append((newValue, sender))
            
        pub.subscribe(onEvent, self.effort.trackingChangedEventType())
        self.effort.setStop(date.DateTime())
        self.assertEqual([(True, self.effort)], events)

    def testNotificationForStopTracking(self):
        self.effort.setStop(date.DateTime())
        events = []
        
        def onEvent(newValue, sender):
            events.append((newValue, sender))

        pub.subscribe(onEvent, self.effort.trackingChangedEventType())
        self.effort.setStop(date.DateTime.now())
        self.assertEqual([(False, self.effort)], events)

    def testRevenueNotificationForTaskHourlyFeeChange(self):
        events = []
        
        def onEvent(newValue, sender):
            events.append((newValue, sender))
            
        pub.subscribe(onEvent, effort.Effort.revenueChangedEventType())
        self.task.setHourlyFee(100)
        self.assertEqual([(2400.0, self.effort)], events)

    def testRevenueNotificationForEffortDurationChange_ChangeStop(self):
        self.task.setHourlyFee(100)
        events = []
        
        def onEvent(newValue, sender):
            events.append((newValue, sender))
            
        pub.subscribe(onEvent, effort.Effort.revenueChangedEventType())
        self.effort.setStop(date.DateTime(2004, 1, 3))
        self.assertEqual([(4800.0, self.effort)], events)

    def testRevenueNotificationForEffortDurationChange_ChangeStart(self):
        self.task.setHourlyFee(100)
        events = []
        
        def onEvent(newValue, sender):
            events.append((newValue, sender))
            
        pub.subscribe(onEvent, effort.Effort.revenueChangedEventType())
        self.effort.setStart(date.DateTime(2004, 1, 1, 12, 0, 0))
        self.assertEqual([(1200.0, self.effort)], events)

    def testDefaultStartAndStop(self):
        effortPeriod = effort.Effort(self.task)
        currentTime = date.DateTime.now()
        now = lambda: currentTime
        self.assertEqual(now() - effortPeriod.getStart(),
            effortPeriod.duration(now=now))
     
    def testState(self):
        state = self.effort.__getstate__()
        theTask = task.Task()
        newEffort = effort.Effort(theTask)
        newEffort.__setstate__(state)
        self.assertEqualEfforts(newEffort, self.effort)
        
    def testCopy(self):
        copyEffort = self.effort.copy()
        self.assertEqualEfforts(copyEffort, self.effort)
        self.assertEqual(copyEffort.description(), 
            self.effort.description())

    def testCopyHasDifferentId(self):
        copyEffort = self.effort.copy()
        self.assertNotEqual(copyEffort.id(), self.effort.id())

    def testCopyHasStatusNew(self):
        self.effort.markDeleted()
        copyEffort = self.effort.copy()
        self.assertEqual(copyEffort.getStatus(), copyEffort.STATUS_NEW)

    def testDescription(self):
        self.effort.setDescription('description')
        self.assertEqual('description', self.effort.description())
        
    def testDescription_Constructor(self):
        newEffort = effort.Effort(self.task, description='description')
        self.assertEqual('description', newEffort.description())
        
    def testSetStop_None(self):
        self.effort.setStop()
        now = date.Now()
        self.failUnless(now - date.ONE_SECOND < self.effort.getStop() < now + date.ONE_SECOND)
        
    def testSetStop_Infinite(self):
        self.effort.setStop(date.DateTime.max)
        self.assertEqual(None, self.effort.getStop())

    def testSetStop_SpecificDateTime(self):
        self.effort.setStop(date.DateTime(2005, 1, 1))
        self.assertEqual(date.DateTime(2005, 1, 1), self.effort.getStop())
        
    def testIsNotBeingTracked_(self): 
        self.failIf(self.effort.isBeingTracked())

    def testIsBeingTracked(self): 
        self.effort.setStop(date.DateTime.max)
        self.failUnless(self.effort.isBeingTracked())
        
    def testSetTaskToNewTaskWillAddItToNewTask(self):
        task2 = task.Task()
        self.effort.setTask(task2)
        self.assertEqual([self.effort], task2.efforts())
        
    def testSetTaskToNewTaskWillRemoveItFromOldTask(self):
        self.task.addEffort(self.effort)
        task2 = task.Task()
        self.effort.setTask(task2)
        self.assertEqual([self.effort], task2.efforts())
        self.failIf(self.effort in self.task.efforts())

    def testSetTaskToOldTaskTwice(self):
        self.task.addEffort(self.effort)
        self.effort.setTask(self.task)
        self.assertEqual([self.effort], self.task.efforts())
        
    def testRevenueWithoutFee(self):
        self.task.addEffort(self.effort)
        self.assertEqual(0, self.effort.revenue())
        
    def testRevenue_HourlyFee(self):
        self.task.setHourlyFee(100)
        self.task.addEffort(self.effort)
        self.assertEqual(self.effort.duration().hours() * 100, 
            self.effort.revenue())
        
    def testRevenue_FixedFee_OneEffort(self):
        self.task.setFixedFee(1000)
        self.task.addEffort(self.effort)
        self.assertEqual(0, self.effort.revenue())

    def testRevenue_FixedFee_OneSmallEffort(self):
        self.task.setFixedFee(1000)
        self.effort.setStop(self.effort.getStart())
        self.assertEqual(0, self.effort.revenue())
        
    def testRevenue_FixedFee_TwoEfforts(self):
        self.task.setFixedFee(1000)
        self.task.addEffort(self.effort)
        self.task.addEffort(effort.Effort(self.task, 
                                          date.DateTime(2005, 1, 1, 10, 0), 
                                          date.DateTime(2005, 1, 1, 22, 0)))
        self.assertEqual(0, self.effort.revenue())

    def testSubject(self):
        self.assertEqual(self.task.subject(), self.effort.subject())
        
    def testNoCategories(self):
        self.assertEqual(self.task.categories(), self.effort.categories())
        
    def testCategories(self):
        self.task.addCategory(category.Category('C'))
        self.assertEqual(self.task.categories(), self.effort.categories())

    def testModificationEventTypes(self):  # pylint: disable=E1003
        self.assertEqual(super(effort.Effort, self.effort).modificationEventTypes() + \
                         [self.effort.taskChangedEventType(), 
                          self.effort.startChangedEventType(), 
                          self.effort.stopChangedEventType()], 
                         self.effort.modificationEventTypes())


class EffortWithoutTaskTest(test.TestCase):   
    def setUp(self):
        self.effort = effort.Effort(None, start=date.DateTime(2005, 1, 1))
        self.task = task.Task()
        self.events = []
        
    def onEvent(self, event):
        self.events.append(event)  # pragma: no cover
        
    def testCreatingAnEffortWithoutTask(self):
        self.assertEqual(None, self.effort.task())
        
    def testSettingTask(self):
        self.effort.setTask(self.task)
        self.assertEqual(self.task, self.effort.task())
    
    def testSettingTask_CausesNoNotification(self):
        patterns.Publisher().registerObserver(self.onEvent, 
                                              self.effort.taskChangedEventType())
        self.effort.setTask(self.task)
        self.failIf(self.events)



'''
Task Coach - Your friendly task manager
Copyright (C) 2011 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import sys, time, wx
from taskcoachlib import operating_system
from ctypes import *


#==============================================================================
# Linux/BSD

if operating_system.isGTK():
    class XScreenSaverInfo(Structure):
        _fields_ = [('window', c_ulong),
                    ('state', c_int),
                    ('kind', c_int),
                    ('til_or_since', c_ulong),
                    ('idle', c_ulong),
                    ('event_mask', c_ulong)]

    class LinuxIdleQuery(object):
        def __init__(self):
            _x11 = CDLL('libX11.so.6')

            self.XOpenDisplay = CFUNCTYPE(c_ulong, c_char_p)(('XOpenDisplay', _x11))
            self.XCloseDisplay = CFUNCTYPE(c_int, c_ulong)(('XCloseDisplay', _x11))
            self.XRootWindow = CFUNCTYPE(c_ulong, c_ulong, c_int)(('XRootWindow', _x11))

            self.dpy = self.XOpenDisplay(None)

            _xss = CDLL('libXss.so.1')

            self.XScreenSaverAllocInfo = CFUNCTYPE(POINTER(XScreenSaverInfo))(('XScreenSaverAllocInfo', _xss))
            self.XScreenSaverQueryInfo = CFUNCTYPE(c_int, c_ulong, c_ulong, POINTER(XScreenSaverInfo))(('XScreenSaverQueryInfo', _xss))

            self.info = self.XScreenSaverAllocInfo()

        def __del__(self):
            self.XCloseDisplay(self.dpy)

        def getIdleSeconds(self):
            self.XScreenSaverQueryInfo(self.dpy, self.XRootWindow(self.dpy, 0), self.info)
            return 1.0 * self.info.contents.idle / 1000

    IdleQuery = LinuxIdleQuery

elif operating_system.isWindows():
    class LASTINPUTINFO(Structure):
        _fields_ = [('cbSize', c_uint), ('dwTime', c_uint)]

    class WindowsIdleQuery(object):
        def __init__(self):
            self.GetTickCount = windll.kernel32.GetTickCount
            self.GetLastInputInfo = windll.user32.GetLastInputInfo

            self.lastInputInfo = LASTINPUTINFO()
            self.lastInputInfo.cbSize = sizeof(self.lastInputInfo)

        def getIdleSeconds(self):
            self.GetLastInputInfo(byref(self.lastInputInfo))
            return (1.0 * self.GetTickCount() - self.lastInputInfo.dwTime) / 1000

    IdleQuery = WindowsIdleQuery

elif operating_system.isMac():
    # When running from source, select the right binary...

    if not hasattr(sys, 'frozen'):
        import struct, os

        if struct.calcsize('L') == 8:
            _subdir = 'ia64'
        else:
            _subdir = 'ia32'

        sys.path.insert(0, os.path.join(os.path.split(__file__)[0],
                                        '..', '..', 'extension', 'macos', 'bin-%s' % _subdir))

    import _idle
    class MacIdleQuery(_idle.Idle):
        def getIdleSeconds(self):
            return self.get()

    IdleQuery = MacIdleQuery


#==============================================================================
#

class IdleNotifier(wx.EvtHandler, IdleQuery):
    STATE_SLEEPING         = 0
    STATE_AWAKE            = 1

    def __init__(self):
        wx.EvtHandler.__init__(self)
        IdleQuery.__init__(self)

        self.state = self.STATE_AWAKE
        self.lastActivity = time.time()
        self.goneToSleep = None

        self._bound = True
        wx.GetApp().Bind(wx.EVT_IDLE, self._OnIdle)

    def stop(self):
        self.pause()

    def pause(self):
        if self._bound:
            wx.GetApp().Unbind(wx.EVT_IDLE, handler=self._OnIdle)
            self._bound = False

    def resume(self):
        self.state = self.STATE_AWAKE
        self.lastActivity = time.time()
        if not self._bound:
            wx.GetApp().Bind(wx.EVT_IDLE, self._OnIdle)

    def _check(self):
        if self.state == self.STATE_AWAKE and time.time() - self.lastActivity >= self.getMinIdleTime():
            self.goneToSleep = self.lastActivity
            self.state = self.STATE_SLEEPING
            self.sleep()
        elif self.state == self.STATE_SLEEPING and time.time() - self.lastActivity < self.getMinIdleTime():
            self.state = self.STATE_AWAKE
            self.wake(self.goneToSleep)

    def _OnIdle(self, event):
        self._check()
        self.lastActivity = time.time() - self.getIdleSeconds()
        self._check()
        event.Skip()

    def poweroff(self):
        """
        Call this when the computer goes to sleep.
        """
        if self._bound:
            wx.GetApp().Unbind(wx.EVT_IDLE, handler=self._OnIdle)
            self._bound = False

    def poweron(self):
        """
        Call this when the computer resumes from sleep.
        """
        if not self._bound:
            wx.GetApp().Bind(wx.EVT_IDLE, self._OnIdle)
            self._bound = True
        self._check()
        self.lastActivity = time.time() - self.getIdleSeconds()
        self._check()

    def getMinIdleTime(self):
        """
        Should return the minimum time in seconds before going idle.
        """
        raise NotImplementedError

    def sleep(self):
        """
        Called when the min idle time has elapsed without any user
        input.
        """

    def wake(self, timestamp):
        """
        Called when the computer is not idle any more.
        """



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''


import wx

from taskcoachlib.i18n import _

class SyncMLWarningDialog(wx.Dialog):
    def __init__(self, parent):
        super(SyncMLWarningDialog, self).__init__(parent, wx.ID_ANY, _('Compatibility warning'))

        textWidget = wx.StaticText(self, wx.ID_ANY,
                                   _('The SyncML feature is disabled, because the module\n'
                                     'could not be loaded. This may be because your platform\n'
                                     'is not supported, or under Windows, you may be missing\n'
                                     'some mandatory DLLs. Please see the SyncML section of\n'
                                     'the online help for details (under "Troubleshooting").'))
        self.checkbox = wx.CheckBox(self, wx.ID_ANY, _('Never show this dialog again'))
        self.checkbox.SetValue(True)
        button = wx.Button(self, wx.ID_ANY, _('OK'))

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(textWidget, 0, wx.ALL, 10)
        sizer.Add(self.checkbox, 0, wx.ALL, 3)
        sizer.Add(button, 0, wx.ALL|wx.ALIGN_CENTRE, 3)

        self.SetSizer(sizer)

        wx.EVT_BUTTON(button, wx.ID_ANY, self.OnOK)
        wx.EVT_CLOSE(self, self.OnOK)

        self.Fit()

    def OnOK(self, event):
        if self.checkbox.IsChecked():
            self.EndModal(wx.ID_OK)
        else:
            self.EndModal(wx.ID_CANCEL)



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib.i18n import _
from taskcoachlib.config import defaults


class TaskStatus(object):
    def __init__(self, statusString, pluralLabel, countLabel, hideMenuText,
                 hideHelpText):
        self.statusString = statusString
        self.pluralLabel = pluralLabel
        self.countLabel = countLabel
        self.hideMenuText = hideMenuText
        self.hideHelpText = hideHelpText

    # This is only used by uicommands so use default if the user configured 'no bitmap', because we
    # need one for the toolbar...

    def getBitmap(self, settings):
        if settings.get('icon', '%stasks' % self.statusString):
            return settings.get('icon', '%stasks' % self.statusString)
        return defaults.defaults['icon']['%stasks' % self.statusString]

    def getHideBitmap(self, settings):
        if settings.get('icon', '%stasks' % self.statusString):
            return '%s+cross_red_icon' % settings.get('icon', '%stasks' % self.statusString)
        return '%s+cross_red_icon' % defaults.defaults['icon']['%stasks' % self.statusString]

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, self.statusString)
        
    def __str__(self):
        return self.statusString
           
    def __eq__(self, other):
        return self.statusString == other.statusString
    
    def __neq__(self, other):
        return self.statusString != other.statusString
    
    def __nonzero__(self):
        return True


inactive = TaskStatus('inactive', _('Inactive tasks'), 
    _('Inactive tasks: %d (%d%%)'), _('Hide &inactive tasks'),
    _('Show/hide inactive tasks (incomplete tasks without actual start date)'))

late = TaskStatus('late', _('Late tasks'), 
    _('Late tasks: %d (%d%%)'), _('Hide &late tasks'), 
    _('Show/hide late tasks (inactive tasks with a planned start in the past)'))

active = TaskStatus('active', _('Active tasks'), 
    _('Active tasks: %d (%d%%)'), _('Hide &active tasks'),
    _('Show/hide active tasks (incomplete tasks with an actual start date in the past)'))

duesoon = TaskStatus('duesoon', _('Due soon tasks'), 
    _('Due soon tasks: %d (%d%%)'), _('Hide &due soon tasks'),
    _('Show/hide due soon tasks (incomplete tasks with a due date in the near future)'))

overdue = TaskStatus('overdue', _('Overdue tasks'), 
    _('Overdue tasks: %d (%d%%)'), _('Hide &over due tasks'),
    _('Show/hide over due tasks (incomplete tasks with a due date in the past)'))

completed = TaskStatus('completed', _('Completed tasks'), 
    _('Completed tasks: %d (%d%%)'), _('Hide &completed tasks'),
    _('Show/hide completed tasks'))



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import wx, test
from taskcoachlib import gui, config, persistence, meta, operating_system
from taskcoachlib.domain import task


class MockViewer(wx.Frame):
    def title(self):
        return ''
    
    def settingsSection(self):
        return 'taskviewer'
    
    def viewerStatusEventType(self):
        return 'mockviewer.status'

    def curselection(self):
        return []


class MainWindowUnderTest(gui.MainWindow):
    def _create_window_components(self):
        # Create only the window components we really need for the tests
        self._create_viewer_container()
        self.viewer.addViewer(MockViewer(None))
        self._create_status_bar()
    

class DummyIOController(object):
    def needSave(self, *args, **kwargs): # pylint: disable=W0613
        return False # pragma: no cover

    def changedOnDisk(self):
        return False # pragme: no cover


class MainWindowTestCase(test.wxTestCase):
    def setUp(self):
        super(MainWindowTestCase, self).setUp()
        self.settings = config.Settings(load=False)
        self.setSettings()
        task.Task.settings = self.settings
        self.taskFile = persistence.TaskFile()
        self.mainwindow = MainWindowUnderTest(DummyIOController(),
            self.taskFile, self.settings)
            
    def setSettings(self):
        pass

    def tearDown(self):
        if operating_system.isMac():
            self.mainwindow.OnQuit() # Stop power monitoring thread
        # Also stop idle time thread
        self.mainwindow._idleController.stop()
        self.mainwindow.Destroy()
        wx.Yield()
        del self.mainwindow
        super(MainWindowTestCase, self).tearDown()
        self.taskFile.close()
        self.taskFile.stop()
        
        
class MainWindowTest(MainWindowTestCase):
    def testStatusBar_Show(self):
        self.settings.setboolean('view', 'statusbar', True)
        self.failUnless(self.mainwindow.GetStatusBar().IsShown())

    def testStatusBar_Hide(self):
        self.settings.setboolean('view', 'statusbar', False)
        self.failIf(self.mainwindow.GetStatusBar().IsShown())

    def testTitle_Default(self):
        self.assertEqual(meta.name, self.mainwindow.GetTitle())
        
    def testTitle_AfterFilenameChange(self):
        self.taskFile.setFilename('New filename')
        self.assertEqual('%s - %s'%(meta.name, self.taskFile.filename()), 
            self.mainwindow.GetTitle())

    def testTitle_AfterChange(self):
        self.taskFile.setFilename('New filename')
        self.taskFile.tasks().extend([task.Task()])
        self.assertEqual('%s - %s *' % (meta.name, self.taskFile.filename()),
                         self.mainwindow.GetTitle())

    def testTitle_AfterSave(self):
        self.taskFile.setFilename('New filename')
        self.taskFile.tasks().extend([task.Task()])
        self.taskFile.save()
        self.assertEqual('%s - %s' % (meta.name, self.taskFile.filename()),
                         self.mainwindow.GetTitle())


class MainWindowMaximizeTestCase(MainWindowTestCase):
    maximized = 'Subclass responsibility'
    
    def setUp(self):
        super(MainWindowMaximizeTestCase, self).setUp()
        if not operating_system.isMac():
            self.mainwindow.Show() # Or IsMaximized() returns always False...
        
    def setSettings(self):
        self.settings.setboolean('window', 'maximized', self.maximized)


class MainWindowNotMaximizedTest(MainWindowMaximizeTestCase):
    maximized = False
    
    def testCreate(self):
        self.failIf(self.mainwindow.IsMaximized())

    @test.skipOnPlatform('__WXGTK__')
    def testMaximize(self): # pragma: no cover
        # Skipping this test under wxGTK. I don't know how it managed
        # to pass before but according to
        # http://trac.wxwidgets.org/ticket/9167 and to my own tests,
        # EVT_MAXIMIZE is a noop under this platform.
        self.mainwindow.Maximize()
        if operating_system.isWindows():
            wx.Yield()
        self.failUnless(self.settings.getboolean('window', 'maximized'))


class MainWindowMaximizedTest(MainWindowMaximizeTestCase):
    maximized = True

    @test.skipOnPlatform('__WXMAC__')
    def testCreate(self):
        self.failUnless(self.mainwindow.IsMaximized()) # pragma: no cover


class MainWindowIconizedTest(MainWindowTestCase):
    def setUp(self):
        super(MainWindowIconizedTest, self).setUp()        
        if operating_system.isGTK():
            wx.SafeYield() # pragma: no cover
            
    def setSettings(self):
        self.settings.set('window', 'starticonized', 'Always')
        
    def expectedHeight(self):
        height = 500
        if operating_system.isMac():
            height += 18 # pragma: no cover
        return height
    
    @test.skipOnPlatform('__WXGTK__') # Test fails on Fedora, don't know why nor how to fix it    
    def testIsIconized(self):
        self.failUnless(self.mainwindow.IsIconized()) # pragma: no cover
                        
    def testWindowSize(self):
        self.assertEqual((900, self.expectedHeight()), 
                         eval(self.settings.get('window', 'size')))
        
    def testWindowSizeShouldnotChangeWhenReceivingChangeSizeEvent(self):
        event = wx.SizeEvent((100, 20))
        process = self.mainwindow.ProcessEvent
        if operating_system.isWindows():
            process(event) # pragma: no cover
        else:
            wx.CallAfter(process, event) # pragma: no cover
        self.assertEqual((900, self.expectedHeight()),
                         eval(self.settings.get('window', 'size')))




'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib import operating_system


class AbstractNotifier(object):
    """
    Abstract base class for interfacing with notification systems
    (Growl, Snarl...).
    """

    notifiers = {}
    _enabled = True

    def getName(self):
        raise NotImplementedError

    def isAvailable(self):
        raise NotImplementedError

    def notify(self, title, summary, bitmap, **kwargs):
        raise NotImplementedError

    @classmethod
    def register(klass, notifier):
        if notifier.isAvailable():
            klass.notifiers[notifier.getName()] = notifier

    @classmethod
    def get(klass, name):
        return klass.notifiers.get(name, None)

    @classmethod
    def getSimple(klass):
        """
        Returns a notifier suitable for simple notifications. This
        defaults to Growl/Snarl/libnotify depending on their
        availability.
        """

        if klass._enabled:
            if operating_system.isMac():
                return klass.get('Growl') or klass.get('Task Coach')
            elif operating_system.isWindows():
                return klass.get('Snarl') or klass.get('Task Coach')
            else:
                return klass.get('libnotify') or klass.get('Task Coach')
        else:
            class DummyNotifier(AbstractNotifier):
                def getName(self):
                    return u'Dummy'
                def isAvailable(self):
                    return True
                def notify(self, title, summary, bitmap, **kwargs):
                    pass
            return DummyNotifier()

    @classmethod
    def disableNotifications(klass):
        klass._enabled = False

    @classmethod
    def names(klass):
        names = klass.notifiers.keys()
        names.sort()
        return names



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib import persistence, config
from taskcoachlib.domain import task, date
import test
import os
from unittests import dummy


class AutoExporterTestCase(test.TestCase):
    def setUp(self):
        task.Task.settings = self.settings = config.Settings(load=False)
        self.exporter = persistence.AutoImporterExporter(self.settings)
        self.taskFile = persistence.TaskFile()
        self.tskFilename = 'autoexport.tsk'
        self.txtFilename = 'autoexport.txt'
        self.taskFile.setFilename(self.tskFilename)
        
    def tearDown(self):
        super(AutoExporterTestCase, self).tearDown()
        del self.exporter
        for filename in self.tskFilename, self.tskFilename + '.delta', self.txtFilename, self.txtFilename + '-meta':
            try:
                os.remove(filename)
            except OSError:
                pass
        
    def testAddOneTaskWhenAutoSaveIsOn(self):
        self.settings.set('file', 'autoexport', '["Todo.txt"]')
        self.settings.set('file', 'autosave', 'True')
        autosaver = persistence.AutoSaver(self.settings)
        theTask = task.Task(subject='Some task')
        self.taskFile.tasks().append(theTask)
        autosaver.on_idle(dummy.Event())
        self.assertEqual('Some task tcid:%s\n' % theTask.id(), file(self.txtFilename, 'r').read())
        
    def testAddOneTaskAndSaveManually(self):
        self.settings.set('file', 'autoexport', '["Todo.txt"]')
        theTask = task.Task(subject='Whatever')
        self.taskFile.tasks().append(theTask)
        self.taskFile.save()
        self.assertEqual('Whatever tcid:%s\n' % theTask.id(), file(self.txtFilename, 'r').read())
        
    def testImportOneTaskWhenSavingManually(self):
        self.settings.set('file', 'autoimport', '["Todo.txt"]')
        with file(self.txtFilename, 'w') as todoTxtFile:
            todoTxtFile.write('Imported task\n')
        self.taskFile.save()
        self.assertEqual('Imported task', 
                         list(self.taskFile.tasks())[0].subject())
        
    def testImportOneTaskWhenAutoSaving(self):
        self.settings.set('file', 'autoimport', '["Todo.txt"]')
        self.settings.set('file', 'autosave', 'True')
        autosaver = persistence.AutoSaver(self.settings)
        with file(self.txtFilename, 'w') as todoTxtFile:
            todoTxtFile.write('Imported task\n')
        self.taskFile.tasks().append(task.Task(subject='Some task'))
        autosaver.on_idle(dummy.Event())
        self.assertEqual(2, len(self.taskFile.tasks()))

    def testImportAfterReadingTaskFile(self):
        self.taskFile.save()
        self.settings.set('file', 'autoimport', '["Todo.txt"]')
        with file(self.txtFilename, 'w') as todoTxtFile:
            todoTxtFile.write('Imported task\n')
        self.taskFile.load()
        self.assertEqual('Imported task', 
                         list(self.taskFile.tasks())[0].subject())
        
    def testSaveWithAutoImportWhenFileToImportDoesNotExist(self):
        self.settings.set('file', 'autoimport', '["Todo.txt"]')
        self.taskFile.tasks().append(task.Task(subject='Whatever'))
        self.taskFile.save()

    def testBothDeletedTask(self):
        self.settings.set('file', 'autoimport', '["Todo.txt"]')
        self.settings.set('file', 'autoexport', '["Todo.txt"]')
        aTask = task.Task(subject='Whatever')
        self.taskFile.tasks().append(aTask)
        self.taskFile.save()
        self.taskFile.tasks().remove(aTask)
        self.taskFile.save()
        self.assertEqual(self.taskFile.tasks(), [])

    def testBothMarkCompleted(self):
        self.settings.set('file', 'autoimport', '["Todo.txt"]')
        self.settings.set('file', 'autoexport', '["Todo.txt"]')
        aTask = task.Task(subject='Whatever')
        self.taskFile.tasks().append(aTask)
        self.taskFile.save()
        now = date.Now()
        aTask.setCompletionDateTime(now)
        self.taskFile.save()
        self.assertEqual(aTask.completionDateTime(), now)



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import StringIO
import test
from taskcoachlib import persistence, gui, config, render
from taskcoachlib.domain import task, effort, date


class CSVWriterTestCase(test.wxTestCase):
    treeMode = 'Subclass responsibility'
    
    def setUp(self):
        super(CSVWriterTestCase, self).setUp()
        task.Task.settings = self.settings = config.Settings(load=False)
        self.fd = StringIO.StringIO()
        self.writer = persistence.CSVWriter(self.fd)
        self.taskFile = persistence.TaskFile()
        self.task = task.Task('Task subject')
        self.taskFile.tasks().append(self.task)
        self.createViewer()

    def tearDown(self):
        super(CSVWriterTestCase, self).tearDown()
        self.taskFile.close()
        self.taskFile.stop()

    def createViewer(self):
        self.settings.set('taskviewer', 'treemode', self.treeMode)
        # pylint: disable=W0201
        self.viewer = gui.viewer.TaskViewer(self.frame, self.taskFile,
            self.settings)

    def __writeAndRead(self, selectionOnly, separateDateAndTimeColumns, columns):
        self.writer.write(self.viewer, self.settings, selectionOnly, 
                          separateDateAndTimeColumns=separateDateAndTimeColumns,
                          columns=columns)
        return self.fd.getvalue()

    def expectInCSV(self, csvFragment, selectionOnly=False, separateDateAndTimeColumns=False,
                    columns=None):
        csv = self.__writeAndRead(selectionOnly, separateDateAndTimeColumns, columns)
        self.failUnless(csvFragment in csv, 
                        '%s not in %s' % (csvFragment, csv))
    
    def expectNotInCSV(self, csvFragment, selectionOnly=False, separateDateAndTimeColumns=False,
                       columns=None):
        csv = self.__writeAndRead(selectionOnly, separateDateAndTimeColumns, columns)
        self.failIf(csvFragment in csv, '%s in %s' % (csvFragment, csv))

    def selectItem(self, items):
        self.viewer.select(items)


class TaskTestsMixin(object):
    def testTaskSubject(self):
        self.expectInCSV('Task subject,')

    def testWriteSelectionOnly(self):
        self.expectNotInCSV('Task subject', selectionOnly=True)
        
    def testWriteSelectionOnly_SelectedChild(self):
        child = task.Task('Child', parent=self.task)
        self.taskFile.tasks().append(child)
        self.viewer.expandAll()
        self.selectItem([child])
        self.expectInCSV('Child,', selectionOnly=True)

    def testWriteSelectionOnly_SelectedParent(self):
        child = task.Task('Child', parent=self.task)
        self.taskFile.tasks().append(child)
        self.selectItem([self.task])
        self.expectNotInCSV('Child', selectionOnly=True)
        
    def testWriteSeparateDateAndTimeColumns(self):
        plannedStartDateTime = date.Now()
        self.task.setPlannedStartDateTime(plannedStartDateTime)
        self.expectInCSV(','.join((render.date(plannedStartDateTime), render.time(plannedStartDateTime))),
                         separateDateAndTimeColumns=True)
        
    def testWriteSeparateDateAndTimeColumnsWithDateBefore1900(self):
        plannedStartDateTime = date.DateTime(1600, 1, 1, 12, 30, 0)
        self.task.setPlannedStartDateTime(plannedStartDateTime)
        self.expectInCSV(','.join((render.date(plannedStartDateTime), render.time(plannedStartDateTime))),
                         separateDateAndTimeColumns=True)       
               
    def testDontWriteSeparateDateAndTimeColumns(self):
        plannedStartDateTime = date.Now()
        self.task.setPlannedStartDateTime(plannedStartDateTime)
        self.expectInCSV(' '.join((render.date(plannedStartDateTime), render.time(plannedStartDateTime))),
                         separateDateAndTimeColumns=False)
                
    def testDontWriteDefaultDateTimes(self):
        defaultDateTime = date.DateTime()
        self.expectNotInCSV(' '.join([render.date(defaultDateTime), render.time(defaultDateTime)]),
                            separateDateAndTimeColumns=False)

    def testDontWriteDefaultDateTimesWithSeparatedDateAndTimeColumns(self):
        defaultDateTime = date.DateTime()
        self.expectNotInCSV(','.join([render.date(defaultDateTime), render.time(defaultDateTime)]),
                            separateDateAndTimeColumns=True)
        
    def testSpecifyColumns(self):
        self.task.setPriority(999)
        self.expectInCSV('999', columns=self.viewer.columns())

    def testPlannedStartDateTimeToday(self):
        today = date.Now()
        self.viewer.showColumnByName('plannedStartDateTime')
        self.task.setPlannedStartDateTime(today)
        self.expectInCSV(render.dateTime(today, humanReadable=False))

    def testPlannedStartDateTimeYesterday(self):
        yesterday = date.Yesterday()
        self.viewer.showColumnByName('plannedStartDateTime')
        self.task.setPlannedStartDateTime(yesterday)
        self.expectInCSV(render.dateTime(yesterday, humanReadable=False))

    def testPlannedStartDateTimeTomorrow(self):
        tomorrow = date.Tomorrow()
        self.viewer.showColumnByName('plannedStartDateTime')
        self.task.setPlannedStartDateTime(tomorrow)
        self.expectInCSV(render.dateTime(tomorrow, humanReadable=False))

    def testPlannedStartDateToday(self):
        today = date.Now().startOfDay()
        self.viewer.showColumnByName('plannedStartDateTime')
        self.task.setPlannedStartDateTime(today)
        self.expectInCSV(render.dateTime(today, humanReadable=False))

    def testPlannedStartDateYesterday(self):
        yesterday = date.Yesterday().startOfDay()
        self.viewer.showColumnByName('plannedStartDateTime')
        self.task.setPlannedStartDateTime(yesterday)
        self.expectInCSV(render.dateTime(yesterday, humanReadable=False))

    def testPlannedStartDateTomorrow(self):
        tomorrow = date.Tomorrow().startOfDay()
        self.viewer.showColumnByName('plannedStartDateTime')
        self.task.setPlannedStartDateTime(tomorrow)
        self.expectInCSV(render.dateTime(tomorrow, humanReadable=False))

    def testDueDateTimeToday(self):
        today = date.Now()
        self.viewer.showColumnByName('dueDateTime')
        self.task.setDueDateTime(today)
        self.expectInCSV(render.dateTime(today, humanReadable=False))

    def testDueDateTimeYesterday(self):
        yesterday = date.Yesterday()
        self.viewer.showColumnByName('dueDateTime')
        self.task.setDueDateTime(yesterday)
        self.expectInCSV(render.dateTime(yesterday, humanReadable=False))

    def testDueDateTimeTomorrow(self):
        tomorrow = date.Tomorrow()
        self.viewer.showColumnByName('dueDateTime')
        self.task.setDueDateTime(tomorrow)
        self.expectInCSV(render.dateTime(tomorrow, humanReadable=False))

    def testDueDateToday(self):
        today = date.Now().startOfDay()
        self.viewer.showColumnByName('dueDateTime')
        self.task.setDueDateTime(today)
        self.expectInCSV(render.dateTime(today, humanReadable=False))

    def testDueDateYesterday(self):
        yesterday = date.Yesterday().startOfDay()
        self.viewer.showColumnByName('dueDateTime')
        self.task.setDueDateTime(yesterday)
        self.expectInCSV(render.dateTime(yesterday, humanReadable=False))

    def testDueDateTomorrow(self):
        tomorrow = date.Tomorrow().startOfDay()
        self.viewer.showColumnByName('dueDateTime')
        self.task.setDueDateTime(tomorrow)
        self.expectInCSV(render.dateTime(tomorrow, humanReadable=False))

    def testActualStartDateTimeToday(self):
        today = date.Now()
        self.viewer.showColumnByName('actualStartDateTime')
        self.task.setActualStartDateTime(today)
        self.expectInCSV(render.dateTime(today, humanReadable=False))

    def testActualStartDateTimeYesterday(self):
        yesterday = date.Yesterday()
        self.viewer.showColumnByName('actualStartDateTime')
        self.task.setActualStartDateTime(yesterday)
        self.expectInCSV(render.dateTime(yesterday, humanReadable=False))

    def testActualStartDateTimeTomorrow(self):
        tomorrow = date.Tomorrow()
        self.viewer.showColumnByName('actualStartDateTime')
        self.task.setActualStartDateTime(tomorrow)
        self.expectInCSV(render.dateTime(tomorrow, humanReadable=False))

    def testActualStartDateToday(self):
        today = date.Now().startOfDay()
        self.viewer.showColumnByName('actualStartDateTime')
        self.task.setActualStartDateTime(today)
        self.expectInCSV(render.dateTime(today, humanReadable=False))

    def testActualStartDateYesterday(self):
        yesterday = date.Yesterday().startOfDay()
        self.viewer.showColumnByName('actualStartDateTime')
        self.task.setActualStartDateTime(yesterday)
        self.expectInCSV(render.dateTime(yesterday, humanReadable=False))

    def testActualStartDateTomorrow(self):
        tomorrow = date.Tomorrow().startOfDay()
        self.viewer.showColumnByName('actualStartDateTime')
        self.task.setActualStartDateTime(tomorrow)
        self.expectInCSV(render.dateTime(tomorrow, humanReadable=False))

    def testCompletionDateTimeToday(self):
        today = date.Now()
        self.viewer.showColumnByName('completionDateTime')
        self.task.setCompletionDateTime(today)
        self.expectInCSV(render.dateTime(today, humanReadable=False))

    def testCompletionDateTimeYesterday(self):
        yesterday = date.Yesterday()
        self.viewer.showColumnByName('completionDateTime')
        self.task.setCompletionDateTime(yesterday)
        self.expectInCSV(render.dateTime(yesterday, humanReadable=False))

    def testCompletionDateTimeTomorrow(self):
        tomorrow = date.Tomorrow()
        self.viewer.showColumnByName('completionDateTime')
        self.task.setCompletionDateTime(tomorrow)
        self.expectInCSV(render.dateTime(tomorrow, humanReadable=False))

    def testCompletionDateToday(self):
        today = date.Now().startOfDay()
        self.viewer.showColumnByName('completionDateTime')
        self.task.setCompletionDateTime(today)
        self.expectInCSV(render.dateTime(today, humanReadable=False))

    def testCompletionDateYesterday(self):
        yesterday = date.Yesterday().startOfDay()
        self.viewer.showColumnByName('completionDateTime')
        self.task.setCompletionDateTime(yesterday)
        self.expectInCSV(render.dateTime(yesterday, humanReadable=False))

    def testCompletionDateTomorrow(self):
        tomorrow = date.Tomorrow().startOfDay()
        self.viewer.showColumnByName('completionDateTime')
        self.task.setCompletionDateTime(tomorrow)
        self.expectInCSV(render.dateTime(tomorrow, humanReadable=False))
        
    def testCreationDateTime(self):
        self.viewer.showColumnByName('creationDateTime')
        self.expectInCSV(render.dateTime(self.task.creationDateTime(), 
                                         humanReadable=False))
        
    def testMissingCreationDateTime(self):
        self.viewer.showColumnByName('creationDateTime')
        self.taskFile.tasks().append(task.Task(creationDateTime=date.DateTime.min))
        self.taskFile.tasks().remove(self.task)
        self.expectInCSV(',,,')  # No 1/1/1 for the missing creation date

    def testModificationDateTime(self):
        self.viewer.showColumnByName('modificationDateTime')
        self.task.setModificationDateTime(date.DateTime(2013, 1, 1, 12, 0, 0))
        self.expectInCSV(render.dateTime(self.task.modificationDateTime(),
                                         humanReadable=False))

    def testMissingModificationDateTime(self):
        self.viewer.showColumnByName('modificationDateTime')
        self.task.setModificationDateTime(date.DateTime.min)
        self.expectInCSV(',,,')  # No 1/1/1 for the missing creation date


class CSVListWriterTest(TaskTestsMixin, CSVWriterTestCase):
    treeMode = 'False'
     
    def testTaskDescription(self):
        self.task.setDescription('Task description')
        self.viewer.showColumnByName('description')
        self.expectInCSV(',Task description,')
 
    def testTaskDescriptionWithNewLine(self):
        self.task.setDescription('Line1\nLine2')
        self.viewer.showColumnByName('description')
        self.expectInCSV('"Line1\nLine2"')
                      

class CSVTreeWriterTest(TaskTestsMixin, CSVWriterTestCase):
    treeMode = 'True'


class EffortWriterTest(CSVWriterTestCase):
    def setUp(self):
        super(EffortWriterTest, self).setUp()
        now = date.DateTime.now()
        self.effort = effort.Effort(self.task, start=now,
                                    stop=now + date.ONE_SECOND)
        self.task.addEffort(self.effort)

    def createViewer(self):
        # pylint: disable=W0201
        self.viewer = gui.viewer.EffortViewer(self.frame, self.taskFile,
            self.settings)

    def testTaskSubject(self):
        self.expectInCSV('Task subject,')
        
    def testEffortDuration(self):
        self.expectInCSV(',0:00:01')
        
    def testEffortPerDay(self):
        self.settings.settext(self.viewer.settingsSection(), 'aggregation',
                              'day')
        self.expectInCSV('Total')

    def testEffortPerDay_SelectionOnly_EmptySelection(self):
        self.settings.settext(self.viewer.settingsSection(), 'aggregation',
                              'day')
        self.expectNotInCSV('Total', selectionOnly=True)

    def testEffortPerDay_SelectionOnly_SelectAll(self):
        self.settings.settext(self.viewer.settingsSection(), 'aggregation',
                              'day')
        self.viewer.widget.select_all()
        self.viewer.updateSelection()
        self.expectInCSV('Total', selectionOnly=True)

    def testExportAllColumns_NoSplit(self):
        self.expectInCSV(render.dateTimePeriod(self.effort.getStart(), self.effort.getStop()),
                         columns=self.viewer.selectableColumns())

    def testExportAllColumns_Split(self):
        self.expectInCSV('%s,%s,%s,%s' % (render.date(self.effort.getStart().date()),
                                          render.time(self.effort.getStart().time()),
                                          render.date(self.effort.getStop().date()),
                                          render.time(self.effort.getStop().time())),
                         separateDateAndTimeColumns=True,
                         columns=self.viewer.selectableColumns())


class EffortWriterRenderTest(CSVWriterTestCase):
    def createViewer(self):
        # pylint: disable=W0201
        self.viewer = gui.viewer.EffortViewer(self.frame, self.taskFile,
            self.settings)

    def testToday(self):
        midnight = date.Now().startOfDay()
        self.task.addEffort(effort.Effort(self.task, start=midnight,
                            stop=midnight + date.TWO_HOURS))
        self.expectNotInCSV('Today')

    def testTomorrow(self):
        midnight = date.Tomorrow().startOfDay()
        self.task.addEffort(effort.Effort(self.task, start=midnight,
                            stop=midnight + date.TWO_HOURS))
        self.expectNotInCSV('Tomorrow')

    def testYesterday(self):
        midnight = date.Yesterday().startOfDay()
        self.task.addEffort(effort.Effort(self.task, start=midnight,
                            stop=midnight + date.TWO_HOURS))
        self.expectNotInCSV('Today')



# -*- coding: utf-8 -*-

'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib import application, meta, widgets, \
    operating_system # pylint: disable=W0622
from taskcoachlib.gui import viewer, toolbar, uicommand, remindercontroller, \
    artprovider, windowdimensionstracker, idlecontroller
from taskcoachlib.gui.dialog.iphone import IPhoneSyncTypeDialog
from taskcoachlib.gui.dialog.xfce4warning import XFCE4WarningDialog
from taskcoachlib.gui.dialog.editor import Editor
from taskcoachlib.gui.iphone import IPhoneSyncFrame
from taskcoachlib.i18n import _
from taskcoachlib.powermgt import PowerStateMixin
from taskcoachlib.help.balloontips import BalloonTipManager
from taskcoachlib.thirdparty.pubsub import pub
import taskcoachlib.thirdparty.aui as aui
import wx, ctypes


def turn_on_double_buffering_on_windows(window):
    # This has actually an adverse effect when Aero is enabled...
    from ctypes import wintypes
    dll = ctypes.WinDLL('dwmapi.dll')
    ret = wintypes.BOOL()
    if dll.DwmIsCompositionEnabled(ctypes.pointer(ret)) == 0 and ret.value:
        return
    import win32gui, win32con  # pylint: disable=F0401
    exstyle = win32gui.GetWindowLong(window.GetHandle(), win32con.GWL_EXSTYLE)
    exstyle |= win32con.WS_EX_COMPOSITED
    win32gui.SetWindowLong(window.GetHandle(), win32con.GWL_EXSTYLE, exstyle)


class MainWindow(PowerStateMixin, BalloonTipManager,
                 widgets.AuiManagedFrameWithDynamicCenterPane):
    def __init__(self, iocontroller, taskFile, settings, *args, **kwargs):
        self.__splash = kwargs.pop('splash', None)
        super(MainWindow, self).__init__(None, -1, '', *args, **kwargs)
        # This prevents the viewers from flickering on Windows 7 when refreshed:
        if operating_system.isWindows7_OrNewer():
            turn_on_double_buffering_on_windows(self)
        self.__dimensions_tracker = windowdimensionstracker.WindowDimensionsTracker(self, settings)
        self.iocontroller = iocontroller
        self.taskFile = taskFile
        self.settings = settings
        self.__filename = None
        self.__dirty = False
        self.__shutdown = False
        self.Bind(wx.EVT_CLOSE, self.onClose)
        self.Bind(wx.EVT_ICONIZE, self.onIconify)
        self.Bind(wx.EVT_SIZE, self.onResize)
        self._create_window_components()  # Not private for test purposes
        self.__init_window_components()
        self.__init_window()
        self.__register_for_window_component_changes()
        
        if settings.getboolean('feature', 'syncml'):
            try:
                import taskcoachlib.syncml.core  # pylint: disable=W0612,W0404
            except ImportError:
                if settings.getboolean('syncml', 'showwarning'):
                    dlg = widgets.SyncMLWarningDialog(self)
                    try:
                        if dlg.ShowModal() == wx.ID_OK:
                            settings.setboolean('syncml', 'showwarning', False)
                    finally:
                        dlg.Destroy()

        self.bonjourRegister = None
        self.bonjourAcceptor = None
        self._registerBonjour()
        pub.subscribe(self._registerBonjour, 'settings.feature.iphone')

        self._idleController = idlecontroller.IdleController(self,
                                                             self.settings,
                                                             self.taskFile.efforts())

        wx.CallAfter(self.checkXFCE4)

    def _registerBonjour(self, value=True):
        if self.bonjourRegister is not None:
            self.bonjourRegister.stop()
            self.bonjourAcceptor.close()
            self.bonjourRegister = self.bonjourAcceptor = None

        if self.settings.getboolean('feature', 'iphone'):
            # pylint: disable=W0612,W0404,W0702
            try:
                from taskcoachlib.thirdparty import pybonjour 
                from taskcoachlib.iphone import IPhoneAcceptor, BonjourServiceRegister

                acceptor = IPhoneAcceptor(self, self.settings, self.iocontroller)
                def success(reader):
                    self.bonjourRegister = reader
                    self.bonjourAcceptor = acceptor
                def error(reason):
                    acceptor.close()
                    wx.MessageBox(reason.getErrorMessage(), _('Error'), wx.OK)
                BonjourServiceRegister(self.settings, acceptor.port).addCallbacks(success, error)
            except:
                from taskcoachlib.gui.dialog.iphone import IPhoneBonjourDialog

                dlg = IPhoneBonjourDialog(self, wx.ID_ANY, _('Warning'))
                try:
                    dlg.ShowModal()
                finally:
                    dlg.Destroy()

    def checkXFCE4(self):
        if operating_system.isGTK():
            mon = application.Application().sessionMonitor
            if mon is not None and \
                    self.settings.getboolean('feature', 'usesm2') and \
                    self.settings.getboolean('feature', 'showsmwarning') and \
                    mon.vendor == 'xfce4-session':
                dlg = XFCE4WarningDialog(self, self.settings)
                dlg.Show()

    def setShutdownInProgress(self):
        self.__shutdown = True

    def _create_window_components(self):  # Not private for test purposes
        self._create_viewer_container()
        viewer.addViewers(self.viewer, self.taskFile, self.settings)
        self._create_status_bar()
        self.__create_menu_bar()
        self.__create_reminder_controller()
        wx.CallAfter(self.viewer.componentsCreated)

    def _create_viewer_container(self):  # Not private for test purposes
        # pylint: disable=W0201
        self.viewer = viewer.ViewerContainer(self, self.settings) 
        
    def _create_status_bar(self):
        from taskcoachlib.gui import status  # pylint: disable=W0404
        self.SetStatusBar(status.StatusBar(self, self.viewer))
        
    def __create_menu_bar(self):
        from taskcoachlib.gui import menu  # pylint: disable=W0404
        self.SetMenuBar(menu.MainMenu(self, self.settings, self.iocontroller, 
                                      self.viewer, self.taskFile))
    
    def __create_reminder_controller(self):
        # pylint: disable=W0201
        self.reminderController = \
            remindercontroller.ReminderController(self, self.taskFile.tasks(),
                self.taskFile.efforts(), self.settings)
        
    def addPane(self, page, caption, floating=False):  # pylint: disable=W0221
        name = page.settingsSection()
        super(MainWindow, self).addPane(page, caption, name, floating=floating)
        
    def __init_window(self):
        self.__filename = self.taskFile.filename()
        self.__setTitle()
        self.SetIcons(artprovider.iconBundle('taskcoach'))
        self.displayMessage(_('Welcome to %(name)s version %(version)s') % \
            {'name': meta.name, 'version': meta.version}, pane=1)

    def __init_window_components(self):
        self.showToolBar(self.settings.getvalue('view', 'toolbar'))
        # We use CallAfter because otherwise the statusbar will appear at the 
        # top of the window when it is initially hidden and later shown.
        wx.CallAfter(self.showStatusBar, 
                     self.settings.getboolean('view', 'statusbar'))
        self.__restore_perspective()
            
    def __restore_perspective(self):
        perspective = self.settings.get('view', 'perspective')
        for viewer_type in viewer.viewerTypes():
            if self.__perspective_and_settings_viewer_count_differ(viewer_type):
                # Different viewer counts may happen when the name of a viewer 
                # is changed between versions
                perspective = ''
                break

        try:
            self.manager.LoadPerspective(perspective)
        except ValueError, reason:
            # This has been reported to happen. Don't know why. Keep going
            # if it does.
            if self.__splash:
                self.__splash.Destroy()
            wx.MessageBox(_('''Couldn't restore the pane layout from TaskCoach.ini:
%s

The default pane layout will be used.

If this happens again, please make a copy of your TaskCoach.ini file '''
'''before closing the program, open a bug report, and attach the '''
'''copied TaskCoach.ini file to the bug report.''') % reason,
            _('%s settings error') % meta.name, style=wx.OK | wx.ICON_ERROR)
            self.manager.LoadPerspective('')
        
        for pane in self.manager.GetAllPanes():
            # Prevent zombie panes by making sure all panes are visible
            if not pane.IsShown():
                pane.Show()
            # Ignore the titles that are saved in the perspective, they may be
            # incorrect when the user changes translation:
            if hasattr(pane.window, 'title'):
                pane.Caption(pane.window.title())
        self.manager.Update()
        
    def __perspective_and_settings_viewer_count_differ(self, viewer_type):
        perspective = self.settings.get('view', 'perspective')
        perspective_viewer_count = perspective.count('name=%s' % viewer_type)
        settings_viewer_count = self.settings.getint('view', 
                                                     '%scount' % viewer_type)
        return perspective_viewer_count != settings_viewer_count
    
    def __register_for_window_component_changes(self):
        pub.subscribe(self.__onFilenameChanged, 'taskfile.filenameChanged')
        pub.subscribe(self.__onDirtyChanged, 'taskfile.dirty')
        pub.subscribe(self.__onDirtyChanged, 'taskfile.clean')
        pub.subscribe(self.showStatusBar, 'settings.view.statusbar')
        pub.subscribe(self.showToolBar, 'settings.view.toolbar')
        self.Bind(aui.EVT_AUI_PANE_CLOSE, self.onCloseToolBar)

    def __onFilenameChanged(self, filename):
        self.__filename = filename
        self.__setTitle()

    def __onDirtyChanged(self, taskFile):
        self.__dirty = taskFile.isDirty()
        self.__setTitle()

    def __setTitle(self):
        title = meta.name
        if self.__filename:
            title += ' - %s' % self.__filename
        if self.__dirty:
            title += ' *'
        self.SetTitle(title)
        
    def displayMessage(self, message, pane=0):
        self.GetStatusBar().SetStatusText(message, pane)
        
    def save_settings(self):
        self.__save_viewer_counts()
        self.__save_perspective()
        self.__save_position()

    def __save_viewer_counts(self):
        ''' Save the number of viewers for each viewer type. '''
        for viewer_type in viewer.viewerTypes():
            count = len([v for v in self.viewer if v.__class__.__name__.lower() == viewer_type])
            self.settings.set('view', viewer_type + 'count', str(count))
            
    def __save_perspective(self):
        perspective = self.manager.SavePerspective()
        self.settings.set('view', 'perspective', perspective)
        
    def __save_position(self):
        self.__dimensions_tracker.save_position()

    def closeEditors(self):
        for child in self.GetChildren():
            if isinstance(child, Editor):
                child.Close()

    def onClose(self, event):
        self.closeEditors()

        if self.__shutdown:
            event.Skip()
            return
        if event.CanVeto() and self.settings.getboolean('window', 
                                                        'hidewhenclosed'):
            event.Veto()
            self.Iconize()
        else:
            if application.Application().quitApplication():
                event.Skip()
                self.taskFile.stop()
                self._idleController.stop()

    def restore(self, event):  # pylint: disable=W0613
        if self.settings.getboolean('window', 'maximized'):
            self.Maximize()
        self.Iconize(False)
        self.Show()
        self.Raise()
        self.Refresh()

    def onIconify(self, event):
        if event.Iconized() and self.settings.getboolean('window', 
                                                         'hidewheniconized'):
            self.Hide()
        else:
            event.Skip()

    def onResize(self, event):
        currentToolbar = self.manager.GetPane('toolbar')
        if currentToolbar.IsOk():
            currentToolbar.window.SetSize((event.GetSize().GetWidth(), -1))
            currentToolbar.window.SetMinSize((event.GetSize().GetWidth(), 42))
        event.Skip()

    def showStatusBar(self, value=True):
        # FIXME: First hiding the statusbar, then hiding the toolbar, then
        # showing the statusbar puts it in the wrong place (only on Linux?)
        self.GetStatusBar().Show(value)
        self.SendSizeEvent()
        
    def createToolBarUICommands(self):
        ''' UI commands to put on the toolbar of this window. ''' 
        uiCommands = [
                uicommand.FileOpen(iocontroller=self.iocontroller), 
                uicommand.FileSave(iocontroller=self.iocontroller),
                uicommand.FileMergeDiskChanges(iocontroller=self.iocontroller),
                uicommand.Print(viewer=self.viewer, settings=self.settings), 
                None, 
                uicommand.EditUndo(), 
                uicommand.EditRedo()]
        uiCommands.extend([ 
            None, 
            uicommand.EffortStartButton(taskList=self.taskFile.tasks()), 
            uicommand.EffortStop(viewer=self.viewer,
                                 effortList=self.taskFile.efforts(),
                                 taskList=self.taskFile.tasks())])
        return uiCommands

    def getToolBarPerspective(self):
        return self.settings.get('view', 'toolbarperspective')

    def saveToolBarPerspective(self, perspective):
        self.settings.set('view', 'toolbarperspective', perspective)

    def showToolBar(self, value):
        currentToolbar = self.manager.GetPane('toolbar')
        if currentToolbar.IsOk():
            self.manager.DetachPane(currentToolbar.window)
            currentToolbar.window.Destroy()
        if value:
            bar = toolbar.MainToolBar(self, self.settings, size=value)
            self.manager.AddPane(bar, aui.AuiPaneInfo().Name('toolbar').
                                 Caption('Toolbar').ToolbarPane().Top().DestroyOnClose().
                                 LeftDockable(False).RightDockable(False))
            # Using .Gripper(False) does not work here
            wx.CallAfter(bar.SetGripperVisible, False)
        self.manager.Update()

    def onCloseToolBar(self, event):
        if event.GetPane().IsToolbar():
            self.settings.setvalue('view', 'toolbar', None)
        event.Skip()
        
    # Viewers
    
    def advanceSelection(self, forward):
        self.viewer.advanceSelection(forward)
        
    def viewerCount(self):
        return len(self.viewer)

    # Power management

    def OnPowerState(self, state):
        pub.sendMessage('powermgt.%s' % {self.POWERON: 'on', self.POWEROFF: 'off'}[state])

    # iPhone-related methods.

    def createIPhoneProgressFrame(self):
        return IPhoneSyncFrame(self.settings, _('iPhone/iPod'),
            icon=wx.ArtProvider.GetBitmap('taskcoach', wx.ART_FRAME_ICON, 
                                          (16, 16)),
            parent=self)

    def getIPhoneSyncType(self, guid):
        if guid == self.taskFile.guid():
            return 0  # two-way

        dlg = IPhoneSyncTypeDialog(self, wx.ID_ANY, _('Synchronization type'))
        try:
            dlg.ShowModal()
            return dlg.value
        finally:
            dlg.Destroy()

    def notifyIPhoneProtocolFailed(self):
        # This should actually never happen.
        wx.MessageBox(_('''An iPhone or iPod Touch device tried to synchronize with this\n'''
                      '''task file, but the protocol negotiation failed. Please file a\n'''
                      '''bug report.'''),
                      _('Error'), wx.OK)

    def clearTasks(self):
        self.taskFile.clear(False)

    def restoreTasks(self, categories, tasks):
        self.taskFile.clear(False)
        self.taskFile.categories().extend(categories)
        self.taskFile.tasks().extend(tasks)

    def addIPhoneCategory(self, category):
        self.taskFile.categories().append(category)

    def removeIPhoneCategory(self, category):
        self.taskFile.categories().remove(category)

    def modifyIPhoneCategory(self, category, name):
        category.setSubject(name)

    def addIPhoneTask(self, task, categories):
        self.taskFile.tasks().append(task)
        for category in categories:
            task.addCategory(category)
            category.addCategorizable(task)

    def removeIPhoneTask(self, task):
        self.taskFile.tasks().remove(task)

    def addIPhoneEffort(self, task, effort):
        if task is not None:
            task.addEffort(effort)

    def modifyIPhoneEffort(self, effort, subject, started, ended):
        effort.setSubject(subject)
        effort.setStart(started)
        effort.setStop(ended)

    def modifyIPhoneTask(self, task, subject, description, plannedStartDateTime, 
                         dueDateTime, completionDateTime, reminderDateTime,
                         recurrence, priority, categories):
        task.setSubject(subject)
        task.setDescription(description)
        task.setPlannedStartDateTime(plannedStartDateTime)
        task.setDueDateTime(dueDateTime)
        task.setCompletionDateTime(completionDateTime)
        task.setReminder(reminderDateTime)
        task.setRecurrence(recurrence)
        task.setPriority(priority)

        if categories is not None:  # Protocol v2
            for toRemove in task.categories() - categories:
                task.removeCategory(toRemove)
                toRemove.removeCategorizable(task)

            for toAdd in categories - task.categories():
                task.addCategory(toAdd)
                toAdd.addCategorizable(task)



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import test
from taskcoachlib import config
from taskcoachlib.domain import task, category

# pylint: disable=W0201,E1101

''' The tests below test the category filter. Different fixtures are defined
    for different combinations of tasks and categories. Each fixture is than
    subclassed by a <Fixture>InListMode and a <Fixture>InTreeMode class since 
    we want to run the tests in both list and tree mode. ''' # pylint: disable=W0105
    

class CategoryFilterHelpersMixin(object):
    def setFilterOnAnyCategory(self):
        self.settings.setboolean('view', 'categoryfiltermatchall', False)
        
    def setFilterOnAllCategories(self):
        self.settings.setboolean('view', 'categoryfiltermatchall', True)

    def link(self, category, categorizable): # pylint: disable=W0621
        category.addCategorizable(categorizable)
        categorizable.addCategory(category)

    def assertChildTaskIsFiltered(self):
        self.assertEqual(set([self.childTask, self.parentTask] if self.treeMode else [self.childTask]), 
                         set(self.filter))
        
    def assertFilterHidesNothing(self):
        self.assertEqual(set(self.tasks), set(self.filter))
        
    def assertFilterHidesEverything(self):
        self.failIf(self.filter)
        

class Fixture(CategoryFilterHelpersMixin):
    treeMode = False
    
    def setUp(self):
        self.settings = task.Task.settings = config.Settings(load=False)
        self.categories = category.CategoryList(self.createCategories())
        self.tasks = task.TaskList(self.createTasks())
        self.categorize()
        self.filter = category.filter.CategoryFilter(self.tasks, 
            categories=self.categories, treeMode=self.treeMode)

    def createTasks(self):
        return []
    
    def createCategories(self):
        return [] # pragma: no cover
    
    def categorize(self):
        pass

    # Common tests that should pass for all fixtures
    
    def testThatFilterContainsAllItemsWhenNotFiltering(self):
        self.assertEqual(self.filter.originalLength(), len(self.filter))

    def testThatFilterOriginalLengthAlwaysEqualsNumberOfTasksWhenNotFiltering(self):
        self.assertEqual(self.filter.originalLength(), len(self.tasks))
        
    def testThatFilterContainsNoItemsWhenRemovingOriginalItems(self):
        self.tasks.clear()
        self.assertEqual(0, self.filter.originalLength())

    def testThatFilterLengthIsSmallerOrEqualThanOriginalLength(self):
        self.failUnless(len(self.filter) <= self.filter.originalLength())
        
    def testThatFilterIsEmptyWhenFilteringOnACategoryWithoutCategorizables(self):
        emptyCategory = category.Category('empty')
        self.categories.append(emptyCategory)
        emptyCategory.setFiltered()
        self.failIf(self.filter)
               
    def testThatFilterContainsAllItemsAfterRemovingFilteredEmptyCategory(self):
        emptyCategory = category.Category('empty')
        self.categories.append(emptyCategory)
        emptyCategory.setFiltered()
        self.categories.remove(emptyCategory)
        self.assertFilterHidesNothing()
    
    def testThatFilterContainsAllItemsAfterFilteringAndUnfiltering(self):
        aCategory = list(self.categories)[0]
        aCategory.setFiltered()
        aCategory.setFiltered(False)
        self.assertFilterHidesNothing()
    
    def testThatFilterContainsAllItemsAfterUnfilteringACategoryThatWasNotFiltered(self):
        aCategory = list(self.categories)[0]
        aCategory.setFiltered(False)
        self.assertFilterHidesNothing()   

    def testThatFilterContainsAllItemsWhenFilteringOnAnyCategoryWithoutAnyCategoryFiltered(self):
        self.setFilterOnAnyCategory()
        self.assertFilterHidesNothing() 

    def testThatFilterContainsAllItemsWhenFilteringOnAllCategoriesWithoutAnyCategoryFiltered(self):
        self.setFilterOnAllCategories()
        self.assertFilterHidesNothing() 
        
    def testThatFilterContainsNewlyAddedTaskThatBelongsToTheFilteredCategory(self):
        aCategory = list(self.categories)[0]
        aCategory.setFiltered()
        newTask = task.Task()
        newTask.addCategory(aCategory)
        self.tasks.append(newTask)
        self.failUnless(newTask in self.filter)
        
    def testThatFilterDoesContainNewlyAddedTaskThatBelongsToTheFilteredCategoryAfterRemoval(self):
        aCategory = list(self.categories)[0]
        aCategory.setFiltered()
        newTask = task.Task()
        newTask.addCategory(aCategory)
        self.tasks.append(newTask)
        self.tasks.remove(newTask)
        self.failIf(newTask in self.filter)
        
    def testThatFilterContainsNoTasksWhenFilteringAllCategoriesIncludingAnEmptyCategory(self):
        emptyCategory = category.Category('empty')
        self.categories.append(emptyCategory)
        for eachCategory in self.categories:
            eachCategory.setFiltered()
        self.setFilterOnAllCategories()        
        self.assertFilterHidesEverything()


class OneCategoryFixture(Fixture):
    def createCategories(self):
        self.category = category.Category('category')
        return [self.category]

    def testThatFilterIsEmptyWhenNoCategoriesAreFiltered(self):
        self.assertEqual(0, len(self.filter))

    def testThatFilterIsEmptyWhenCategoryIsFiltered(self):
        self.category.setFiltered()
        self.assertEqual(0, len(self.filter))
                

class OneCategoryFilterInListModeTest(OneCategoryFixture, test.TestCase):
    treeMode = False   
    

class OneCategoryFilterInTreeModeTest(OneCategoryFixture, test.TestCase):
    treeMode = True


class OneCategoryAndOneTaskFixture(Fixture):
    def createCategories(self):
        self.category = category.Category('category')
        return [self.category]
        
    def createTasks(self):
        self.task = task.Task('task')
        return [self.task]

    def testThatFilterContainsUncategorizedTaskWhenNoCategoriesAreFiltered(self):
        self.assertFilterHidesNothing()

    def testThatFilterDoesNotContainUncategorizedTaskWhenCategoryIsFiltered(self):
        self.category.setFiltered()
        self.assertFilterHidesEverything()
        
    def testThatFilterContainsCategorizedTaskWhenCategoryIsFiltered(self):
        self.link(self.category, self.task)
        self.category.setFiltered()
        self.assertFilterHidesNothing()


class OneCategoryAndOneTaskInListModeTest(OneCategoryAndOneTaskFixture, test.TestCase):
    treeMode = False   
    

class OneCategoryAndOneTaskInTreeModeTest(OneCategoryAndOneTaskFixture, test.TestCase):
    treeMode = True


class OneCategoryAndTwoTasksFixture(Fixture):
    def createCategories(self):
        self.category = category.Category('category')
        return [self.category]
        
    def createTasks(self):
        self.task1 = task.Task('task1')
        self.task2 = task.Task('task2')
        return [self.task1, self.task2]

    def testThatFilterContainsAllUncategorizedTaskWhenCategoryIsNotFiltered(self):
        self.assertFilterHidesNothing()

    def testThatFilterContainsNoUncategorizedTasksWhenCategoryIsFiltered(self):
        self.category.setFiltered()
        self.assertFilterHidesEverything()
        
    def testThatFilterContainsCategorizedTaskWhenCategoryIsFiltered(self):
        self.category.setFiltered()
        self.link(self.category, self.task1)
        self.assertEqual([self.task1], list(self.filter))

    def testThatFilterContainsCategorizedTasksWhenCategoryIsFiltered(self):
        self.category.setFiltered()
        self.link(self.category, self.task1)
        self.link(self.category, self.task2)
        self.assertFilterHidesNothing()


class OneCategoryAndTwoTasksInListModeTest(OneCategoryAndTwoTasksFixture, test.TestCase):
    treeMode = False   
    

class OneCategoryAndTwoTasksInTreeModeTest(OneCategoryAndTwoTasksFixture, test.TestCase):
    treeMode = True


class TwoCategoriesAndOneTaskFixture(Fixture):
    def createCategories(self):
        self.category1 = category.Category('category1')
        self.category2 = category.Category('category2')
        return [self.category1, self.category2]
        
    def createTasks(self):
        self.task = task.Task('task')
        return [self.task]
    
    def categorize(self):
        self.link(self.category1, self.task)
    
    def testThatFilterContainsTaskWhenFilteringOnAnyCategoryAndBothAreFiltered(self):
        self.setFilterOnAnyCategory()
        self.category1.setFiltered()
        self.category2.setFiltered()
        self.assertFilterHidesNothing()

    def testThatFilterDoesNotContainTaskWhenFilteringOnAllCategoriesAndBothAreFiltered(self):
        self.setFilterOnAllCategories()
        self.category1.setFiltered()
        self.category2.setFiltered()
        self.assertFilterHidesEverything()


class TwoCategoriesAndOneTaskInListModeTest(TwoCategoriesAndOneTaskFixture, test.TestCase):
    treeMode = False   
    

class TwoCategoriesAndOneTaskInTreeModeTest(TwoCategoriesAndOneTaskFixture, test.TestCase):
    treeMode = True
        

class TwoCategoriesAndTwoTasksFixture(Fixture):
    def createCategories(self):
        self.category1 = category.Category('category1')
        self.category2 = category.Category('category2')
        return [self.category1, self.category2]
        
    def createTasks(self):
        self.task1 = task.Task('task1')
        self.task2 = task.Task('task2')
        return [self.task1, self.task2]

    def testThatFilterContainsAllUncategorizedTaskWhenCategoryIsNotFiltered(self):
        self.assertFilterHidesNothing()

    def testThatFilterContainsNoUncategorizedTasksWhenCategoryIsFiltered(self):
        self.category1.setFiltered()
        self.assertFilterHidesEverything()
        
    def testThatFilterContainsNoUncategorizedTasksWhenAnyCategoryIsFiltered(self):
        self.setFilterOnAnyCategory()
        self.category1.setFiltered()
        self.category2.setFiltered()
        self.assertFilterHidesEverything()
        
    def testThatFilterContainsNoUncategorizedTasksWhenAllCategoriesAreFiltered(self):
        self.setFilterOnAllCategories()
        self.category1.setFiltered()
        self.category2.setFiltered()
        self.assertFilterHidesEverything()
        
    def testThatFilterContainsCategorizedTaskWhenAnyCategoryIsFiltered(self):
        self.link(self.category1, self.task1)
        self.setFilterOnAnyCategory()
        self.category1.setFiltered()
        self.category2.setFiltered()
        self.assertEqual([self.task1], list(self.filter))

    def testThatFilterDoesNotContainCategorizedTaskWhenAllCategoriesAreFiltered(self):
        self.link(self.category1, self.task1)
        self.setFilterOnAllCategories()
        self.category1.setFiltered()
        self.category2.setFiltered()
        self.assertFilterHidesEverything()
        
    def testThatFilterContainsCategorizedTaskWhenFilteringByThatCategory(self):
        self.link(self.category1, self.task1)
        self.category1.setFiltered()
        self.assertEqual([self.task1], list(self.filter))

    def testThatFilterDoesNotContainCategorizedTaskWhenFilteringByAnotherCategory(self):
        self.link(self.category1, self.task1)
        self.category2.setFiltered()
        self.assertFilterHidesEverything()

    def testThatFilterContainsCategorizedTasksWhenFilteringByThatCategory(self):
        self.category1.setFiltered()
        self.link(self.category1, self.task1)
        self.link(self.category1, self.task2)
        self.assertFilterHidesNothing()

    def testThatFilterContainsCategorizedTasksWhenFilteringByAnyCategory(self):
        self.setFilterOnAnyCategory()
        self.category1.setFiltered()
        self.category2.setFiltered()
        self.link(self.category1, self.task1)
        self.link(self.category1, self.task2)
        self.assertFilterHidesNothing()

    def testThatFilterDoesNotContainCategorizedTasksWhenFilteringByAllCategories(self):
        self.setFilterOnAllCategories()
        self.category1.setFiltered()
        self.category2.setFiltered()
        self.link(self.category1, self.task1)
        self.link(self.category1, self.task2)
        self.assertFilterHidesEverything()


class TwoCategoriesAndTwoTasksInListModeTest(TwoCategoriesAndTwoTasksFixture, test.TestCase):
    treeMode = False   
    

class TwoCategoriesAndTwoTasksInTreeModeTest(TwoCategoriesAndTwoTasksFixture, test.TestCase):
    treeMode = True


class OneCategoryAndParentAndChildTaskFixture(Fixture):
    def createCategories(self):
        self.category = category.Category('category')
        return [self.category]
        
    def createTasks(self):
        self.parentTask = task.Task('parent')
        self.childTask = task.Task('child')
        self.parentTask.addChild(self.childTask)
        self.childTask.setParent(self.parentTask)
        return [self.parentTask, self.childTask]

    def testThatFilterContainsChildWhenParentIsCategorizedAndFiltered(self):
        self.link(self.category, self.parentTask)
        self.category.setFiltered()
        self.assertFilterHidesNothing()

    def testThatFilterContainsParentWhenChildIsCategorizedAndFilteredButOnlyInTreeMode(self):
        self.link(self.category, self.childTask)
        self.category.setFiltered()
        self.assertEqual(2 if self.treeMode else 1, len(self.filter))
        

class OneCategoryAndParentAndChildTaskInListModeTest(OneCategoryAndParentAndChildTaskFixture, test.TestCase):
    treeMode = False   
    

class OneCategoryAndParentAndChildTaskInTreeModeTest(OneCategoryAndParentAndChildTaskFixture, test.TestCase):
    treeMode = True


class TwoCategoriesAndParentAndChildTaskFixture(Fixture):
    def createCategories(self):
        self.category1 = category.Category('category1')
        self.category2 = category.Category('category2')
        return [self.category1, self.category2]
        
    def createTasks(self):
        self.parentTask = task.Task('parent')
        self.childTask = task.Task('child')
        self.parentTask.addChild(self.childTask)
        self.childTask.setParent(self.parentTask)
        return [self.parentTask, self.childTask]
    
    def categorize(self):
        self.link(self.category1, self.parentTask)
        self.link(self.category2, self.childTask)

    def testThatFilterContainsBothParentAndChildTaskWhenFilteringOnAnyCategoryAndBothCategories(self):
        self.setFilterOnAnyCategory()
        self.category1.setFiltered()
        self.category2.setFiltered()
        self.assertFilterHidesNothing()

    def testThatFilterContainsBothParentAndChildTaskWhenFilteringOnAllCategoriesAndBothCategories(self):
        self.setFilterOnAllCategories()
        self.category1.setFiltered()
        self.category2.setFiltered()
        self.assertChildTaskIsFiltered()
        

class TwoCategoriesAndParentAndChildTaskInListModeTest(TwoCategoriesAndParentAndChildTaskFixture, test.TestCase):
    treeMode = False   
    

class TwoCategoriesAndParentAndChildTaskInTreeModeTest(TwoCategoriesAndParentAndChildTaskFixture, test.TestCase):
    treeMode = True


class ParentAndChildCategoryAndParentAndChildTaskFixture(Fixture):
    def createCategories(self):
        self.parentCategory = category.Category('parent')
        self.childCategory = category.Category('child')
        self.parentCategory.addChild(self.childCategory)
        self.childCategory.setParent(self.parentCategory)
        return [self.parentCategory, self.childCategory]
        
    def createTasks(self):
        self.parentTask = task.Task('parent')
        self.childTask = task.Task('child')
        self.parentTask.addChild(self.childTask)
        self.childTask.setParent(self.parentTask)
        return [self.parentTask, self.childTask]
    
    def testThatFilterContainsBothTasksWhenParentTaskIsInParentCategoryAndFilteringParentCategory(self):
        self.link(self.parentCategory, self.parentTask)
        self.parentCategory.setFiltered()
        self.assertFilterHidesNothing()

    def testThatFilterContainsNoTasksWhenParentTaskIsInParentCategoryAndFilteringChildCategory(self):
        self.link(self.parentCategory, self.parentTask)
        self.childCategory.setFiltered()
        self.assertFilterHidesEverything()

    def testThatFilterContainsBothTasksWhenParentTaskIsInChildCategoryAndFilteringParentCategory(self):
        self.link(self.childCategory, self.parentTask)
        self.parentCategory.setFiltered()
        self.assertFilterHidesNothing()

    def testThatFilterContainsBothTasksWhenParentTaskIsInChildCategoryAndFilteringChildCategory(self):
        self.link(self.childCategory, self.parentTask)
        self.childCategory.setFiltered()
        self.assertFilterHidesNothing()

    def testThatFilterContainsChildTaskWhenChildTaskIsInParentCategoryAndFilteringParentCategory(self):
        self.link(self.parentCategory, self.childTask)
        self.parentCategory.setFiltered()
        self.assertChildTaskIsFiltered()

    def testThatFilterContainsNoTasksWhenChildTaskIsInParentCategoryAndFilteringChildCategory(self):
        self.link(self.parentCategory, self.childTask)
        self.childCategory.setFiltered()
        self.assertFilterHidesEverything()

    def testThatFilterContainsChildTaskWhenChildTaskIsInChildCategoryAndFilteringParentCategory(self):
        self.link(self.childCategory, self.childTask)
        self.parentCategory.setFiltered()
        self.assertChildTaskIsFiltered()

    def testThatFilterContainsChildTaskWhenChildTaskIsInChildCategoryAndFilteringChildCategory(self):
        self.link(self.childCategory, self.childTask)
        self.childCategory.setFiltered()
        self.assertChildTaskIsFiltered()

    def testThatFilterContainsBothTasksWhenParentTaskIsInChildCategoryAndFilteringBothCategories(self):
        self.setFilterOnAllCategories()
        self.link(self.childCategory, self.parentTask)
        self.parentCategory.setFiltered()
        self.childCategory.setFiltered()
        self.assertFilterHidesNothing()

    def testThatFilterContainsChildTaskWhenChildTaskIsInChildCategoryAndFilteringBothCategories(self):
        self.setFilterOnAllCategories()
        self.link(self.childCategory, self.childTask)
        self.parentCategory.setFiltered()
        self.childCategory.setFiltered()
        self.assertChildTaskIsFiltered()
        
        
class ParentAndChildCategoryAndParentAndChildTaskInListModeTest(ParentAndChildCategoryAndParentAndChildTaskFixture, test.TestCase):
    treeMode = False   
    

class ParentAndChildCategoryAndParentAndChildTaskInTreeModeTest(ParentAndChildCategoryAndParentAndChildTaskFixture, test.TestCase):
    treeMode = True
    

class ParentAndChildCategoryAndParentAndGrandChildTaskFixture(Fixture):
    def createCategories(self):
        self.parentCategory = category.Category('parent')
        self.childCategory = category.Category('child')
        self.parentCategory.addChild(self.childCategory)
        self.childCategory.setParent(self.parentCategory)
        return [self.parentCategory, self.childCategory]
        
    def createTasks(self):
        self.parentTask = task.Task('parent')
        self.childTask = task.Task('child')
        self.grandChildTask = task.Task('grandchild')
        self.parentTask.addChild(self.childTask)
        self.childTask.setParent(self.parentTask)
        self.childTask.addChild(self.grandChildTask)
        self.grandChildTask.setParent(self.childTask)
        return [self.parentTask, self.childTask, self.grandChildTask]

    def testThatFilterContainsAllTasksWhenParentTaskIsInChildCategoryAndFiltered(self):
        self.link(self.childCategory, self.parentTask)
        self.parentCategory.setFiltered()
        self.assertFilterHidesNothing()

    def testThatFilterContainsAllTasksWhenParentTaskIsInChildCategoryAndAllCategoriesAreFiltered(self):
        self.setFilterOnAllCategories()
        self.link(self.childCategory, self.parentTask)
        self.parentCategory.setFiltered()
        self.childCategory.setFiltered()
        self.assertFilterHidesNothing()

    def testThatFilterContainsGrandChildTaskWhenChildTaskIsInChildCategoryAndParentCategoryIsFiltered(self):
        self.link(self.childCategory, self.childTask)
        self.parentCategory.setFiltered()
        self.failUnless(self.grandChildTask in self.filter)

    def testThatFilterContainsGrandChildTaskWhenChildTaskIsInChildCategoryAndAllCategoriesAreFiltered(self):
        self.setFilterOnAllCategories()
        self.link(self.childCategory, self.childTask)
        self.parentCategory.setFiltered()
        self.childCategory.setFiltered()
        self.failUnless(self.grandChildTask in self.filter)
        
    def testThatFilterContainsGrandParentWhenGrandChildIsInChildCategoryAndParentCategoryIsFiltered(self):
        self.link(self.childCategory, self.grandChildTask)
        self.parentCategory.setFiltered()
        self.assertEqual(set(self.tasks) if self.treeMode else set([self.grandChildTask]), 
                         set(self.filter))
        
    
class ParentAndChildCategoryAndParentAndGrandChildTaskInListModeTest(ParentAndChildCategoryAndParentAndGrandChildTaskFixture, test.TestCase):
    treeMode = False   
    

class ParentAndChildCategoryAndParentAndGrandChildTaskInTreeModeTest(ParentAndChildCategoryAndParentAndGrandChildTaskFixture, test.TestCase):
    treeMode = True


class TwoCategoriesAndParentAndGrandChildTaskFixture(Fixture):
    def createCategories(self):
        self.category1 = category.Category('category1')
        self.category2 = category.Category('category2')
        return [self.category1, self.category2]
        
    def createTasks(self):
        self.parentTask = task.Task('parent')
        self.childTask = task.Task('child')
        self.grandChildTask = task.Task('grandchild')
        self.parentTask.addChild(self.childTask)
        self.childTask.setParent(self.parentTask)
        self.childTask.addChild(self.grandChildTask)
        self.grandChildTask.setParent(self.childTask)
        return [self.parentTask, self.childTask, self.grandChildTask]
    
    def categorize(self):
        self.link(self.category1, self.parentTask)
        self.link(self.category2, self.grandChildTask)

    def testThatFilterContainsGrandChildWhenFilteringOnAllCategories(self):
        self.setFilterOnAllCategories()
        self.category1.setFiltered()
        self.category2.setFiltered()
        self.assertEqual(set(self.tasks) if self.treeMode else set([self.grandChildTask]), 
                         set(self.filter))

    def testThatFilterContainsGrandChildWhenFilteringOnAnyCategory(self):
        self.setFilterOnAnyCategory()
        self.category1.setFiltered()
        self.category2.setFiltered()
        self.assertFilterHidesNothing()
        

class TwoCategoriesAndParentAndGrandChildTaskInListModeTest(TwoCategoriesAndParentAndGrandChildTaskFixture, test.TestCase):
    treeMode = False   
    

class TwoCategoriesAndParentAndGrandChildTaskInTreeModeTest(TwoCategoriesAndParentAndGrandChildTaskFixture, test.TestCase):
    treeMode = True

    
class TwoCategoriesAndParentWithTwoChildTasksFixture(Fixture):
    def createCategories(self):
        self.category1 = category.Category('category1')
        self.category2 = category.Category('category2')
        return [self.category1, self.category2]
        
    def createTasks(self):
        self.parentTask = task.Task('parent')
        self.child1Task = task.Task('child1')
        self.child2Task = task.Task('child2')
        self.parentTask.addChild(self.child1Task)
        self.parentTask.addChild(self.child2Task)
        self.child1Task.setParent(self.parentTask)
        self.child2Task.setParent(self.parentTask)
        return [self.parentTask, self.child1Task, self.child2Task]
    
    def categorize(self):
        self.link(self.category1, self.child1Task)
        self.link(self.category2, self.child2Task)

    def testThatFilterIsEmptyWhenFilteringOnAllCategoriesAndChildTasksHaveTwoDifferentCategories(self):
        self.setFilterOnAllCategories()
        self.category1.setFiltered()
        self.category2.setFiltered()
        self.assertFilterHidesEverything()

    def testThatFilterContainsAllTasksWhenFilteringOnAnyCategoryAndChildTasksHaveTwoDifferentCategories(self):
        self.setFilterOnAnyCategory()
        self.category1.setFiltered()
        self.category2.setFiltered()
        self.assertEqual(3 if self.treeMode else 2, len(self.filter))

    def testThatFilterHidesUnfilteredChild(self):
        self.category1.setFiltered()
        self.failIf(self.child2Task in self.filter)

        
class TwoCategoriesAndParentWithTwoChildTasksInListModeTest(TwoCategoriesAndParentWithTwoChildTasksFixture, test.TestCase):
    treeMode = False   
    

class TwoCategoriesAndParentWithTwoChildTasksInTreeModeTest(TwoCategoriesAndParentWithTwoChildTasksFixture, test.TestCase):
    treeMode = True


class ParentAndChildCategoryAndOneTaskFixture(Fixture):
    def createCategories(self):
        self.parentCategory = category.Category('parent')
        self.childCategory = category.Category('child')
        self.parentCategory.addChild(self.childCategory)
        self.childCategory.setParent(self.parentCategory)
        return [self.parentCategory, self.childCategory]
        
    def createTasks(self):
        self.task = task.Task('task')
        return [self.task]

    def testThatFilterDoesNotContainTaskIfTaskHasChildCategoryAndParentIsFiltered(self):
        self.link(self.parentCategory, self.task)
        self.childCategory.setFiltered()
        self.failIf(self.filter)

    def testThatFilterContainsTaskIfTaskIfTaskHasParentCategoryAndChildIsFiltered(self):
        self.link(self.childCategory, self.task)
        self.parentCategory.setFiltered()
        self.assertFilterHidesNothing()

    def testThatFilterContainsTaskIfTaskHasParentAndChildCategoryAndAnyCategoryIsFiltered(self):
        self.setFilterOnAnyCategory()
        self.link(self.parentCategory, self.task)
        self.link(self.childCategory, self.task)
        self.parentCategory.setFiltered()
        self.childCategory.setFiltered()
        self.assertFilterHidesNothing()
        
    def testThatFilterContainsTaskIfTaskHasParentAndChildCategoryAndAllCategoriesAreFiltered(self):
        self.setFilterOnAllCategories()
        self.link(self.parentCategory, self.task)
        self.link(self.childCategory, self.task)
        self.parentCategory.setFiltered()
        self.childCategory.setFiltered()
        self.assertFilterHidesNothing()


class ParentAndChildCategoryAndTaskInListModeTest(ParentAndChildCategoryAndOneTaskFixture, test.TestCase):
    treeMode = False   
    

class ParentAndChildCategoryAndTaskInTreeModeTest(ParentAndChildCategoryAndOneTaskFixture, test.TestCase):
    treeMode = True

                       
class CategoryFilterAndViewFilterFixtureAndCommonTestsMixin(CategoryFilterHelpersMixin):
    def setUp(self):
        task.Task.settings = config.Settings(load=False)
        self.parent = task.Task('parent task')
        self.parent.setShouldMarkCompletedWhenAllChildrenCompleted(False)
        self.child = task.Task('child task')
        self.child.setCompletionDateTime()
        self.childCategory = category.Category('child category')
        self.childCategory.addCategorizable(self.child)
        self.parent.addChild(self.child)
        self.tasks = task.TaskList([self.parent, self.child])
        self.categories = category.CategoryList([self.childCategory])
        self.viewFilter = task.filter.ViewFilter(self.tasks, treeMode=self.treeMode) 
        self.categoryFilter = category.filter.CategoryFilter(self.viewFilter, 
            categories=self.categories, treeMode=self.treeMode)

    def testThatParentIsHiddenWhenHiddenCompletedChildIsFiltered(self):
        self.viewFilter.hideTaskStatus(task.status.completed)
        self.assertEqual(1, len(self.viewFilter))
        self.childCategory.setFiltered(True)
        self.assertEqual(0, len(self.categoryFilter))
        
    def testThatParentIsShownWhenHiddenCompletedChildIsUnfiltered(self):
        self.viewFilter.hideTaskStatus(task.status.completed)
        self.childCategory.setFiltered(True)
        self.assertEqual(0, len(self.categoryFilter))
        self.childCategory.setFiltered(False)
        self.assertEqual(1, len(self.categoryFilter))
        
    def testThatParentIsHiddenWhenFilteredCompletedChildIsHidden(self):
        self.childCategory.setFiltered(True)
        self.assertEqual(2, len(self.viewFilter))
        self.viewFilter.hideTaskStatus(task.status.completed)
        self.assertEqual(0, len(self.categoryFilter))        
        
    def testThatParentIsShownWhenFilteredCompletedChildIsUnhidden(self):
        self.childCategory.setFiltered(True)
        self.viewFilter.hideTaskStatus(task.status.completed)
        self.assertEqual(0, len(self.categoryFilter))
        self.viewFilter.hideTaskStatus(task.status.completed, False)
        self.assertEqual(2 if self.treeMode else 1, len(self.categoryFilter))


class CategoryFilterAndViewFilterInListModeTest(CategoryFilterAndViewFilterFixtureAndCommonTestsMixin, 
                                                test.TestCase):
    treeMode = False   


class CategoryFilterAndViewFilterInTreeModeTest(CategoryFilterAndViewFilterFixtureAndCommonTestsMixin, 
                                                test.TestCase):
    treeMode = True



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import test
from taskcoachlib.gui.dialog import entry 
from taskcoachlib.domain import date
from taskcoachlib import config


class DateTimeEntryTest(test.wxTestCase):
    def setUp(self):
        super(DateTimeEntryTest, self).setUp()
        self.dateTimeEntry = entry.DateTimeEntry(self.frame, 
                                                 config.Settings(load=False))
        self.dateTime = date.DateTime(2004, 1, 1)

    def testCreate(self):
        self.assertEqual(date.DateTime(), self.dateTimeEntry.GetValue())

    def testSet(self):
        now = date.Now()
        self.dateTimeEntry.SetValue(now)
        self.assertAlmostEqual(now.toordinal(), 
                               self.dateTimeEntry.GetValue().toordinal(), places=2)

    def testReset(self):
        self.dateTimeEntry.SetValue()
        self.assertEqual(date.DateTime(), self.dateTimeEntry.GetValue())

    def testValidDateTime(self):
        self.dateTimeEntry.SetValue(self.dateTime)
        self.assertEqual(self.dateTime, self.dateTimeEntry.GetValue())


class DateEntryConstructorTest(test.wxTestCase):
    def testCreateWithDate(self):
        tomorrow = date.Tomorrow()
        dateTimeEntry = entry.DateTimeEntry(self.frame, 
                                            config.Settings(load=False), 
                                            tomorrow)
        self.assertAlmostEqual(tomorrow.toordinal(), 
                               dateTimeEntry.GetValue().toordinal(),
                               places=2)


class TimeDeltaEntryTest(test.wxTestCase):
    def setUp(self):
        super(TimeDeltaEntryTest, self).setUp()
        self.timeDeltaEntry = entry.TimeDeltaEntry(self.frame)
        
    def testDefaultValue(self):
        self.assertEqual(date.TimeDelta(), self.timeDeltaEntry.GetValue())
        
    def testDefaultDisplayedValue(self):    
        self.assertEqual('        0:00:00', 
                         self.timeDeltaEntry._entry.GetValue())
        
    def testSetValue(self):
        self.timeDeltaEntry.SetValue(date.TimeDelta(hours=10, seconds=5))
        self.assertEqual('       10:00:05', 
                         self.timeDeltaEntry._entry.GetValue())
    
    def testOverflow(self):
        self.timeDeltaEntry.SetValue(date.TimeDelta(hours=12345678912))
        self.assertEqual('123456789:00:00', 
                         self.timeDeltaEntry._entry.GetValue())
    
    
class ReadOnlyTimeDeltaEntryTest(test.wxTestCase):
    def setUp(self):
        super(ReadOnlyTimeDeltaEntryTest, self).setUp()
        self.timeDeltaEntry = entry.TimeDeltaEntry(self.frame, readonly=True)

    def testSetNegativeValue(self):
        self.timeDeltaEntry.SetValue(date.TimeDelta(hours=-10, minutes=-20))
        self.assertEqual('      -10:20:00', 
                         self.timeDeltaEntry._entry.GetValue())

    def testSetSmallNegativeValue(self):
        self.timeDeltaEntry.SetValue(date.TimeDelta(seconds=-4))
        self.assertEqual('       -0:00:04', 
                         self.timeDeltaEntry._entry.GetValue())



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from __future__ import absolute_import # For xml...

import os, shutil, glob, math, re
from taskcoachlib.domain import date
from taskcoachlib.thirdparty.pubsub import pub
import bz2, hashlib

# Hack: indirect 
from xml.etree import ElementTree as ET


def SHA(filename):
    return hashlib.sha1(filename.encode('UTF-8')).hexdigest()


def compressFile(srcName, dstName):
    with file(srcName, 'rb') as src:
        dst = bz2.BZ2File(dstName, 'w')
        try:
            shutil.copyfileobj(src, dst)
        finally:
            dst.close()


class BackupManifest(object):
    def __init__(self, settings):
        self.__settings = settings

        xmlName = os.path.join(settings.pathToBackupsDir(), 'backups.xml')
        if os.path.exists(xmlName):
            with file(xmlName, 'rb') as fp:
                root = ET.parse(fp).getroot()
                self.__files = dict([(node.attrib['sha'], node.text) for node in root.findall('file')])
        else:
            self.__files = dict()

    def save(self):
        root = ET.Element('backupfiles')
        for sha, filename in self.__files.items():
            node = ET.SubElement(root, 'file')
            node.attrib['sha'] = sha
            node.text = filename
        with file(os.path.join(self.__settings.pathToBackupsDir(), 'backups.xml'), 'wb') as fp:
            ET.ElementTree(root).write(fp)

    def listFiles(self):
        return sorted(self.__files.values())

    def listBackups(self, filename):
        backups = list()
        for name in os.listdir(self.backupPath(filename)):
            try:
                comp = map(int, [name[0:4], name[4:6], name[6:8], name[8:10], name[10:12], name[12:14]])
            except:
                continue
            backups.append(date.DateTime(*tuple(comp)))
        return list(reversed(sorted(backups)))

    def hasBackups(self, filename):
        return len(self.listBackups(filename)) != 0

    def backupPath(self, filename):
        path = os.path.join(self.__settings.pathToBackupsDir(), SHA(filename))
        if not os.path.exists(path):
            os.makedirs(path)
        return path

    def addFile(self, filename):
        self.__files[SHA(filename)] = filename

    def removeFile(self, filename):
        sha = SHA(filename)
        if sha in self.__files:
            del self.__files[sha]

    def restoreFile(self, filename, dateTime, dstName):
        if os.path.exists(dstName):
            os.remove(dstName)
        sha = SHA(filename)
        src = bz2.BZ2File(os.path.join(self.__settings.pathToBackupsDir(), sha, dateTime.strftime('%Y%m%d%H%M%S.bak')), 'r')
        try:
            with file(dstName, 'wb') as dst:
                shutil.copyfileobj(src, dst)
        finally:
            src.close()


class AutoBackup(object):
    ''' AutoBackup creates a backup copy of the task
        file before it is overwritten. To prevent the number of backups growing
        indefinitely, AutoBackup removes older backups. '''

    minNrOfBackupFiles = 3 # Keep at least three backup files.
    maxNrOfBackupFilesToRemoveAtOnce = 3 # Slowly reduce the number of backups

    def __init__(self, settings, copyfile=compressFile):
        super(AutoBackup, self).__init__()
        self.__settings = settings
        self.__copyfile = copyfile
        pub.subscribe(self.onTaskFileAboutToSave, 'taskfile.aboutToSave')
        pub.subscribe(self.onTaskFileRead, 'taskfile.justRead')

    def onTaskFileRead(self, taskFile):
        ''' Copies old-style backups (in the same dictory as the task file) to the
        user-specific backup directory. The backup directory layout is as follows:

          <backupdir>/backups.xml          List of backups
          <backupdir>/<sha>/<datetime>.bak Backup for <datetime>. <sha> is the SHA-1
                                           hash of the task file name.

        backups.xml maps the SHA to actual file names, for enumeration in the
        GUI. '''

        if not taskFile.filename():
            return

        # First add the file to the XML manifest.
        man = BackupManifest(self.__settings)
        man.addFile(taskFile.filename())
        man.save()

        # Then copy existing backups
        rx = re.compile(r'\.(\d{8})-(\d{6})\.tsk\.bak$')
        for name in os.listdir(os.path.split(taskFile.filename())[0] or '.'):
            try:
                srcName = os.path.join(os.path.split(taskFile.filename())[0], name)
            except UnicodeDecodeError:
                # See https://bugs.debian.org/cgi-bin/bugreport.cgi?bug=764763
                continue

            mt = rx.search(name)
            if mt:
                dstName = os.path.join(man.backupPath(taskFile.filename()), '%s%s.bak' % (mt.group(1), mt.group(2)))
                if os.path.exists(dstName):
                    os.remove(dstName)
                with file(srcName, 'rb') as src:
                    dst = bz2.BZ2File(dstName, 'w')
                    try:
                        shutil.copyfileobj(src, dst)
                    finally:
                        dst.close()
                os.remove(srcName)

    def onTaskFileAboutToSave(self, taskFile):
        ''' Just before a task file is about to be saved, and backups are on,
            create a backup and remove extraneous backup files. '''
        if taskFile.exists():
            self.createBackup(taskFile)
            self.removeExtraneousBackupFiles(taskFile)

    def createBackup(self, taskFile):
        filename = self.backupFilename(taskFile)
        path = os.path.dirname(filename)
        if not os.path.exists(path):
            os.makedirs(path)
        self.__copyfile(taskFile.filename(), filename)

    def removeExtraneousBackupFiles(self, taskFile, remove=os.remove,
                                    glob=glob.glob): # pylint: disable=W0621
        backupFiles = self.backupFiles(taskFile, glob)
        for _ in range(min(self.maxNrOfBackupFilesToRemoveAtOnce,
                           self.numberOfExtraneousBackupFiles(backupFiles))):
            try:
                remove(self.leastUniqueBackupFile(backupFiles))
            except OSError:
                pass # Ignore errors

    def numberOfExtraneousBackupFiles(self, backupFiles):
        return max(0, len(backupFiles) - self.maxNrOfBackupFiles(backupFiles))

    def maxNrOfBackupFiles(self, backupFiles):
        ''' The maximum number of backup files we keep depends on the age of
            the oldest backup file. The older the oldest backup file (that is
            never removed), the more backup files we keep. '''
        if not backupFiles:
            return 0
        age = date.DateTime.now() - self.backupDateTime(backupFiles[0])
        ageInMinutes = age.hours() * 60
        # We keep log(ageInMinutes) backups, but at least minNrOfBackupFiles:
        return max(self.minNrOfBackupFiles, int(math.log(max(1, ageInMinutes))))

    def leastUniqueBackupFile(self, backupFiles):
        ''' Find the backupFile that is closest (in time) to its neighbors,
            i.e. that is the least unique. Ignore the oldest and newest
            backups. '''
        assert len(backupFiles) > self.minNrOfBackupFiles
        deltas = []
        for index in range(1, len(backupFiles)-1):
            delta = self.backupDateTime(backupFiles[index+1]) - \
                    self.backupDateTime(backupFiles[index-1])
            deltas.append((delta, backupFiles[index]))
        deltas.sort()
        return deltas[0][1]

    def backupFiles(self, taskFile, glob=glob.glob):  # pylint: disable=W0621
        sha = SHA(taskFile.filename())
        root = os.path.join(self.__settings.pathToBackupsDir(), sha)
        return sorted(glob('%s.bak' % os.path.join(root, '[0-9]' * 14)))

    def backupFilename(self, taskFile, now=date.DateTime.now):
        ''' Generate a backup filename for the specified date/time. '''
        sha = SHA(taskFile.filename())
        return os.path.join(self.__settings.pathToBackupsDir(), sha, now().strftime('%Y%m%d%H%M%S.bak'))

    @staticmethod
    def backupDateTime(backupFilename):
        ''' Parse the date and time from the filename and return a DateTime
            instance. '''
        dt = os.path.split(backupFilename)[-1][:-4]
        parts = (int(part) for part in (dt[0:4], dt[4:6], dt[6:8],
                                        dt[8:10], dt[10:12], dt[12:14]))
        return date.DateTime(*parts) # pylint: disable=W0142
